import ast
import importlib
import io
import json
import logging
import math

import numpy as np
import os
import random
import string
import matplotlib.pyplot as plt
import time
import traceback
import warnings
from collections import defaultdict, deque
from hashlib import sha256
from tempfile import gettempdir
from types import FunctionType, ModuleType

import numpy as np
import portalocker
from scipy.misc import imsave

try:
    import torch
except ImportError as e:
    import warnings
    warnings.warn(ImportWarning("Could not import Pytorch related modules:\n%s"
        % e.msg))


    class torch:
        dtype = None


class CustomJSONEncoder(json.JSONEncoder):

    def _encode(self, obj):
        raise NotImplementedError

    def _encode_switch(self, obj):
        if isinstance(obj, list):
            return [self._encode_switch(item) for item in obj]
        elif isinstance(obj, dict):
            return {self._encode_key(key): self._encode_switch(val) for key, val in obj.items()}
        else:
            return self._encode(obj)

    def _encode_key(self, obj):
        return self._encode(obj)

    def encode(self, obj):
        return super(CustomJSONEncoder, self).encode(self._encode_switch(obj))

    def iterencode(self, obj, *args, **kwargs):
        return super(CustomJSONEncoder, self).iterencode(self._encode_switch(obj), *args, **kwargs)


class MultiTypeEncoder(CustomJSONEncoder):

    def _encode_key(self, obj):
        if isinstance(obj, int):
            return "__int__({})".format(obj)
        elif isinstance(obj, float):
            return "__float__({})".format(obj)
        else:
            return self._encode(obj)

    def _encode(self, obj):
        if isinstance(obj, tuple):
            return "__tuple__({})".format(obj)
        elif isinstance(obj, np.integer):
            return "__int__({})".format(obj)
        elif isinstance(obj, np.floating):
            return "__float__({})".format(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return obj


class ModuleMultiTypeEncoder(MultiTypeEncoder):

    def _encode(self, obj, strict=False):
        if type(obj) == type:
            return "__type__({}.{})".format(obj.__module__, obj.__name__)
        elif type(obj) == torch.dtype:
            return "__type__({})".format(str(obj))
        elif isinstance(obj, FunctionType):
            return "__function__({}.{})".format(obj.__module__, obj.__name__)
        elif isinstance(obj, ModuleType):
            return "__module__({})".format(obj.__name__)
        else:
            try:
                return super(ModuleMultiTypeEncoder, self)._encode(obj)
            except Exception as e:
                if strict:
                    raise e
                else:
                    message = "Could not pickle object of type {}\n".format(type(obj))
                    message += traceback.format_exc()
                    warnings.warn(message)
                    return repr(obj)


class CustomJSONDecoder(json.JSONDecoder):

    def _decode(self, obj):
        raise NotImplementedError

    def _decode_switch(self, obj):
        if isinstance(obj, list):
            return [self._decode_switch(item) for item in obj]
        elif isinstance(obj, dict):
            return {self._decode_key(key): self._decode_switch(val) for key, val in obj.items()}
        else:
            return self._decode(obj)

    def _decode_key(self, obj):
        return self._decode(obj)

    def decode(self, obj):
        return self._decode_switch(super(CustomJSONDecoder, self).decode(obj))


class MultiTypeDecoder(CustomJSONDecoder):

    def _decode(self, obj):
        if isinstance(obj, str):
            if obj.startswith("__int__"):
                return int(obj[8:-1])
            elif obj.startswith("__float__"):
                return float(obj[10:-1])
            elif obj.startswith("__tuple__"):
                return tuple(ast.literal_eval(obj[10:-1]))
        return obj


class ModuleMultiTypeDecoder(MultiTypeDecoder):

    def _decode(self, obj):
        if isinstance(obj, str):
            if obj.startswith("__type__"):
                str_ = obj[9:-1]
                module_ = ".".join(str_.split(".")[:-1])
                name_ = str_.split(".")[-1]
                type_ = str_
                try:
                    type_ = getattr(importlib.import_module(module_), name_)
                except Exception as e:
                    warnings.warn("Could not load {}".format(str_))
                return type_
            elif obj.startswith("__function__"):
                str_ = obj[13:-1]
                module_ = ".".join(str_.split(".")[:-1])
                name_ = str_.split(".")[-1]
                type_ = str_
                try:
                    type_ = getattr(importlib.import_module(module_), name_)
                except Exception as e:
                    warnings.warn("Could not load {}".format(str_))
                return type_
            elif obj.startswith("__module__"):
                str_ = obj[11:-1]
                type_ = str_
                try:
                    type_ = importlib.import_module(str_)
                except Exception as e:
                    warnings.warn("Could not load {}".format(str_))
                return type_
        return super(ModuleMultiTypeDecoder, self)._decode(obj)


class StringMultiTypeDecoder(CustomJSONDecoder):

    def _decode(self, obj):
        if isinstance(obj, str):
            if obj.startswith("__int__"):
                return obj[8:-1]
            elif obj.startswith("__float__"):
                return obj[10:-1]
            elif obj.startswith("__tuple__"):
                return obj[10:-1]
            elif obj.startswith("__type__"):
                return obj[9:-1]
            elif obj.startswith("__function__"):
                return obj[13:-1]
            elif obj.startswith("__module__"):
                return obj[11:-1]
        return obj


class Singleton:
    """
    A non-thread-safe helper class to ease implementing singletons.
    This should be used as a decorator -- not a metaclass -- to the
    class that should be a singleton.

    The decorated class can define one `__init__` function that
    takes only the `self` argument. Also, the decorated class cannot be
    inherited from. Other than that, there are no restrictions that apply
    to the decorated class.

    To get the singleton instance, use the `Instance` method. Trying
    to use `__call__` will result in a `TypeError` being raised.

    """

    _instance = None

    def __init__(self, decorated):
        self._decorated = decorated

    def get_instance(self, **kwargs):
        """
        Returns the singleton instance. Upon its first call, it creates a
        new instance of the decorated class and calls its `__init__` method.
        On all subsequent calls, the already created instance is returned.

        """
        if not self._instance:
            self._instance = self._decorated(**kwargs)
            return self._instance
        else:
            return self._instance

    def __call__(self):
        raise TypeError('Singletons must be accessed through `get_instance()`.')
        # return self.get_instance()

    def __instancecheck__(self, inst):
        return isinstance(inst, self._decorated)


def get_image_as_buffered_file(image_array):
    """
    Returns a images as file pointer in a buffer

    Args:
        image_array: (C,W,H) To be returned as a file pointer

    Returns:
        Buffer file-pointer object containing the image file
    """
    buf = io.BytesIO()
    imsave(name=buf, arr=image_array.transpose((1, 2, 0)), format="png")
    buf.seek(0)

    return buf


def figure_to_image(figures, close=True):
    """Render matplotlib figure to numpy format.

    Note that this requires the ``matplotlib`` package.
    (https://tensorboardx.readthedocs.io/en/latest/_modules/tensorboardX/utils.html#figure_to_image)

    Args:
        figure (matplotlib.pyplot.figure) or list of figures: figure or a list of figures
        close (bool): Flag to automatically close the figure

    Returns:
        numpy.array: image in [CHW] order
    """
    try:
        import matplotlib.pyplot as plt
        import matplotlib.backends.backend_agg as plt_backend_agg
    except ModuleNotFoundError:
        print('please install matplotlib')

    def render_to_rgb(figure):
        canvas = plt_backend_agg.FigureCanvasAgg(figure)
        canvas.draw()
        data = np.frombuffer(canvas.buffer_rgba(), dtype=np.uint8)
        w, h = figure.canvas.get_width_height()
        image_hwc = data.reshape([h, w, 4])[:, :, 0:3]
        image_chw = np.moveaxis(image_hwc, source=2, destination=0)
        if close:
            plt.close(figure)
        return image_chw

    if isinstance(figures, list):
        images = [render_to_rgb(figure) for figure in figures]
        return np.stack(images)
    else:
        image = render_to_rgb(figures)
        return image


def savefig_and_close(figure, filename, close=True):
    fig_img = figure_to_image(figure, close=close)
    imsave(filename, np.transpose(fig_img, (1, 2, 0)))


def random_string(length):
    random.seed()
    return "".join(random.choice(string.ascii_letters + string.digits) for _ in range(length))


def create_folder(path):
    """
    Creates a folder if not already exists
    Args:
        :param path: The folder to be created
    Returns
        :return: True if folder was newly created, false if folder already exists
    """

    if not os.path.exists(path):
        os.makedirs(path)
        return True
    else:
        return False


def name_and_iter_to_filename(name, n_iter, ending, iter_format="{:05d}", prefix=False):
    iter_str = iter_format.format(n_iter)
    if prefix:
        name = iter_str + "_" + name + ending
    else:
        name = name + "_" + iter_str + ending

    return name


class SafeDict(dict):
    def __missing__(self, key):
        return "{" + key + "}"


class PyLock(object):
    def __init__(self, name, timeout, check_interval=0.25):
        self._timeout = timeout
        self._check_interval = check_interval

        lock_directory = gettempdir()
        unique_token = sha256(name.encode()).hexdigest()
        self._filepath = os.path.join(lock_directory, 'ilock-' + unique_token + '.lock')

    def __enter__(self):

        current_time = call_time = time.time()
        while call_time + self._timeout > current_time:
            self._lockfile = open(self._filepath, 'w')
            try:
                portalocker.lock(self._lockfile, portalocker.constants.LOCK_NB | portalocker.constants.LOCK_EX)
                return self
            except portalocker.exceptions.LockException:
                pass

            current_time = time.time()
            check_interval = self._check_interval if self._timeout > self._check_interval else self._timeout
            time.sleep(check_interval)

        raise RuntimeError('Timeout was reached')

    def __exit__(self, exc_type, exc_val, exc_tb):
        portalocker.unlock(self._lockfile)
        self._lockfile.close()


class LogDict(dict):
    def __init__(self, file_name, base_dir=None, to_console=False, mode="a"):
        """Initializes a new Dict which can log to a given target file."""

        super(LogDict, self).__init__()

        self.file_name = file_name
        if base_dir is not None:
            self.file_name = os.path.join(base_dir, file_name)

        self.logging_identifier = random_string(15)
        self.logger = logging.getLogger("logdict-" + self.logging_identifier)
        self.logger.setLevel(logging.INFO)
        file_handler_formatter = logging.Formatter('')

        self.file_handler = logging.FileHandler(self.file_name, mode=mode)
        self.file_handler.setFormatter(file_handler_formatter)
        self.logger.addHandler(self.file_handler)
        self.logger.propagate = to_console

    def __setitem__(self, key, item):
        super(LogDict, self).__setitem__(key, item)

    def log_complete_content(self):
        """Logs the current content of the dict to the output file as a whole."""
        self.logger.info(str(self))


class ResultLogDict(LogDict):
    def __init__(self, file_name, base_dir=None, running_mean_length=10, **kwargs):
        """Initializes a new Dict which directly logs value changes to a given target_file."""
        super(ResultLogDict, self).__init__(file_name=file_name, base_dir=base_dir, **kwargs)

        self.is_init = False
        self.running_mean_dict = defaultdict(lambda: deque(maxlen=running_mean_length))

        self.__cntr_dict = defaultdict(float)

        if self.file_handler.mode == "w" or os.stat(self.file_handler.baseFilename).st_size == 0:
            self.print_to_file("[")

        self.is_init = True

    def __setitem__(self, key, item):

        if key == "__cntr_dict":
            raise ValueError("In ResultLogDict you can not add an item with key '__cntr_dict'")

        data = item
        if isinstance(item, dict) and "data" in item and "label" in item and "epoch" in item:

            data = item["data"]
            if "counter" in item and item["counter"] is not None:
                self.__cntr_dict[key] = item["counter"]
            json_dict = {key: ResultElement(data=data, label=item["label"], epoch=item["epoch"],
                                            counter=self.__cntr_dict[key])}
        else:
            json_dict = {key: ResultElement(data=data, counter=self.__cntr_dict[key])}
        self.__cntr_dict[key] += 1
        self.logger.info(json.dumps(json_dict) + ",")

        self.running_mean_dict[key].append(data)

        super(ResultLogDict, self).__setitem__(key, data)

    def print_to_file(self, text):
        self.logger.info(text)

    def load(self, reload_dict):
        for key, item in reload_dict.items():

            if isinstance(item, dict) and "data" in item and "label" in item and "epoch" in item:
                data = item["data"]
                if "counter" in item and item["counter"] is not None:
                    self.__cntr_dict[key] = item["counter"]
            else:
                data = item
            self.__cntr_dict[key] += 1

            super(ResultLogDict, self).__setitem__(key, data)

    def close(self):

        self.file_handler.close()
        # Remove trailing comma, unless we've only written "[".
        # This approach (fixed offset) sometimes fails upon errors and the like,
        # we could alternatively read the whole file,
        # parse to only keep "clean" rows and rewrite.
        with open(self.file_handler.baseFilename, "rb+") as handle:
            if os.stat(self.file_handler.baseFilename).st_size > 2:
                handle.seek(-2, os.SEEK_END)
                handle.truncate()
        with open(self.file_handler.baseFilename, "a") as handle:
            handle.write("\n]")


class ResultElement(dict):
    def __init__(self, data=None, label=None, epoch=None, counter=None):
        super(ResultElement, self).__init__()

        if data is not None:
            if isinstance(data, np.floating):
                data = float(data)
            if isinstance(data, np.integer):
                data = int(data)
            self["data"] = data
        if label is not None:
            self["label"] = label
        if epoch is not None:
            self["epoch"] = epoch
        if counter is not None:
            self["counter"] = counter


def chw_to_hwc(np_array):
    if len(np_array.shape) != 3:
        return np_array
    elif np_array.shape[0] != 1 and np_array.shape[0] != 3:
        return np_array
    elif np_array.shape[2] == 1 or np_array.shape[2] == 3:
        return np_array
    else:
        np_array = np.transpose(np_array, (1, 2, 0))
        return np_array


def np_make_grid(np_array, nrow=8, padding=2,
                 normalize=False, range=None, scale_each=False, pad_value=0, to_int=False):
    """Make a grid of images.

    Args:
        np_array (numpy array): 4D mini-batch Tensor of shape (B x C x H x W)
            or a list of images all of the same size.
        nrow (int, optional): Number of images displayed in each row of the grid.
            The Final grid size is (B / nrow, nrow). Default is 8.
        padding (int, optional): amount of padding. Default is 2.
        normalize (bool, optional): If True, shift the image to the range (0, 1),
            by subtracting the minimum and dividing by the maximum pixel value.
        range (tuple, optional): tuple (min, max) where min and max are numbers,
            then these numbers are used to normalize the image. By default, min and max
            are computed from the tensor.
        scale_each (bool, optional): If True, scale each image in the batch of
            images separately rather than the (min, max) over all images.
        pad_value (float, optional): Value for the padded pixels.
        to_int (bool): Transforms the np array to a unit8 array with min 0 and max 255

    Example:
        See this notebook `here <https://gist.github.com/anonymous/bf16430f7750c023141c562f3e9f2a91>`_

    """
    if not (isinstance(np_array, np.ndarray) or
            (isinstance(np_array, list) and all(isinstance(a, np.ndarray) for a in np_array))):
        raise TypeError('Numpy array or list of tensors expected, got {}'.format(type(np_array)))

    # if list of tensors, convert to a 4D mini-batch Tensor
    if isinstance(np_array, list):
        np_array = np.stack(np_array, axis=0)

    if len(np_array.shape) == 2:  # single image H x W
        np_array = np_array.reshape((1, np_array.shape[0], np_array.shape[1]))
    if len(np_array.shape) == 3:  # single image
        if np_array.shape[0] == 1:  # if single-channel, convert to 3-channel
            np_array = np.concatenate((np_array, np_array, np_array), 0)
        np_array = np_array.reshape((1, np_array.shape[0], np_array.shape[1], np_array.shape[2]))

    if len(np_array.shape) == 3 == 4 and np_array.shape[1] == 1:  # single-channel images
        np_array = np.concatenate((np_array, np_array, np_array), 1)

    if normalize is True:
        np_array = np.copy(np_array)  # avoid modifying tensor in-place
        if range is not None:
            assert isinstance(range, tuple), \
                "range has to be a tuple (min, max) if specified. min and max are numbers"

        def norm_ip(img, min_, max_):
            img = np.clip(img, a_min=min_, a_max=max_)
            img = (img - min_) / (max_ - min_ + 1e-5)
            return img

        def norm_range(t, range_=None):
            if range_ is not None:
                t = norm_ip(t, range_[0], range_[1])
            else:
                t = norm_ip(t, float(t.min()), float(t.max()))
            return t

        if scale_each is True:
            for i in np.arange(np_array.shape[0]):  # loop over mini-batch dimension
                np_array[i] = norm_range(np_array[i], range)
        else:
            np_array = norm_range(np_array, range)

    if np_array.shape[0] == 1:
        return np_array.squeeze(0)

    # make the mini-batch of images into a grid
    nmaps = np_array.shape[0]
    xmaps = min(nrow, nmaps)
    ymaps = int(math.ceil(float(nmaps) / xmaps))
    height, width = int(np_array.shape[2] + padding), int(np_array.shape[3] + padding)
    grid = np.zeros((3, height * ymaps + padding, width * xmaps + padding))
    grid += pad_value
    k = 0
    for y in np.arange(ymaps):
        for x in np.arange(xmaps):
            if k >= nmaps:
                break
            grid[:,
            y * height + padding: y * height + padding + height - padding,
            x * width + padding: x * width + padding + width - padding] = np_array[k]
            k = k + 1

    if to_int:
        grid = np.clip(grid * 255, a_min=0, a_max=255)
        grid = grid.astype(np.uint8)

    return grid


def get_tensor_embedding(tensor, method="tsne", n_dims=2, n_neigh=30, **meth_args):
    """
    Return a embedding of a tensor (in a lower dimensional space, e.g. t-SNE)

    Args:
       tensor: Tensor to be embedded
       method: Method used for embedding, options are: tsne, standard, ltsa, hessian, modified, isomap, mds,
       spectral, umap
       n_dims: dimensions to embed the data into
       n_neigh: Neighbour parameter to kind of determin the embedding (see t-SNE for more information)
       **meth_args: Further arguments which can be passed to the embedding method

    Returns:
        The embedded tensor

    """
    from sklearn import manifold
    import umap

    linears = ['standard', 'ltsa', 'hessian', 'modified']
    if method in linears:

        loclin = manifold.LocallyLinearEmbedding(n_neigh, n_dims, method=method, **meth_args)
        emb_data = loclin.fit_transform(tensor)

    elif method == "isomap":
        iso = manifold.Isomap(n_neigh, n_dims, **meth_args)
        emb_data = iso.fit_transform(tensor)

    elif method == "mds":
        mds = manifold.MDS(n_dims, **meth_args)
        emb_data = mds.fit_transform(tensor)

    elif method == "spectral":
        se = manifold.SpectralEmbedding(n_components=n_dims, n_neighbors=n_neigh, **meth_args)
        emb_data = se.fit_transform(tensor)

    elif method == "tsne":
        tsne = manifold.TSNE(n_components=n_dims, perplexity=n_neigh, **meth_args)
        emb_data = tsne.fit_transform(tensor)

    elif method == "umap":
        um = umap.UMAP(n_components=n_dims, n_neighbors=n_neigh, **meth_args)
        emb_data = um.fit_transform(tensor)

    else:
        emb_data = tensor

    return emb_data



from trixi.logger.file.numpyplotfilelogger import NumpyPlotFileLogger
from trixi.logger.file.textfilelogger import TextFileLogger

try:
    from trixi.logger.file.pytorchplotfilelogger import PytorchPlotFileLogger
except ImportError as e:
    import warnings
    warnings.warn(ImportWarning("Could not import Pytorch related modules:\n%s"
        % e.msg))



import argparse
import json
import os
import sys
from collections import OrderedDict

import colorlover as cl
from flask import Blueprint, Flask, abort, render_template, request

from trixi.experiment_browser.dataprocessing import group_images, make_graphs, merge_results, process_base_dir
from trixi.experiment_browser.experimentreader import ExperimentReader, CombiExperimentReader, group_experiments_by
from trixi.util import Config

# These keys will be ignored when in a config file
IGNORE_KEYS = ("name",
               "experiment_dir",
               "work_dir",
               "config_dir",
               "log_dir",
               "checkpoint_dir",
               "img_dir",
               "plot_dir",
               "save_dir",
               "result_dir",
               "time",
               "state")

# Set the color palette for plots
COLORMAP = cl.scales["8"]["qual"]["Dark2"]


def parse_args():
    # Read in base directory

    if len(sys.argv) > 1:

        parser = argparse.ArgumentParser()
        parser.add_argument("base_directory",
                            help="Give the path to the base directory of your project files",
                            type=str, default=".")
        parser.add_argument("-d", "--debug", action="store_true",
                            help="Turn debug mode on, eg. for live reloading.")
        parser.add_argument("-x", "--expose", action="store_true",
                            help="Make server externally visible")
        parser.add_argument("-p", "--port", default=5000, type=int,
                            help="Port to start the server on (5000 by default)")
        args = parser.parse_args()
        base_dir = args.base_directory
        if base_dir[-1] == os.sep:
            base_dir = base_dir[:-1]

    else:
        args = type('', (), {})()
        args.port = 5000
        args.expose = False
        args.debug = False
        args.base_directory = "."
        base_dir = "."

    return args, base_dir


def create_flask_app(base_dir):
    # The actual flask app lives in the package directory. The blueprint allows us
    # to specify an additional static folder and we use that to give access to the
    # experiment files
    app = Flask(__name__, static_folder=os.path.join(os.path.dirname(__file__), "static"))
    blueprint = Blueprint("data", __name__, static_url_path=base_dir, static_folder=base_dir)
    app.register_blueprint(blueprint)

    return app


def register_url_routes(app, base_dir):
    app.add_url_rule("/", "overview", lambda: overview(base_dir), methods=["GET"])
    app.add_url_rule("/overview", "overview_", lambda: overview_(base_dir), methods=["GET"])
    app.add_url_rule('/experiment', "experiment", lambda: experiment(base_dir), methods=['GET'])
    app.add_url_rule('/combine', "combine", lambda: combine(base_dir), methods=['GET'])
    app.add_url_rule('/experiment_log', "experiment_log", lambda: experiment_log(base_dir), methods=['GET'])
    app.add_url_rule('/experiment_plots', "experiment_plots", lambda: experiment_plots(base_dir), methods=['GET'])
    app.add_url_rule('/experiment_remove', "experiment_remove", lambda: experiment_remove(base_dir), methods=['GET'])
    app.add_url_rule('/experiment_star', "experiment_star", lambda: experiment_star(base_dir), methods=['GET'])
    app.add_url_rule('/experiment_rename', "experiment_rename", lambda: experiment_rename(base_dir), methods=['GET'])


def start_browser():
    args, base_dir = parse_args()

    base_dir = os.path.abspath(base_dir)

    app = create_flask_app(base_dir)
    register_url_routes(app, base_dir)

    host = "0.0.0.0" if args.expose else "localhost"
    app.run(debug=args.debug, host=host, port=args.port)


def overview(base_dir):
    try:
        base_info = process_base_dir(base_dir, ignore_keys=IGNORE_KEYS)
        base_info["title"] = base_dir
        return render_template("overview.html", **base_info)
    except Exception as e:
        print(e.__repr__())
        raise e
        abort(500)


def overview_(base_dir):
    dir_ = request.args.get("dir")
    try:
        base_info = process_base_dir(base_dir, dir_, ignore_keys=IGNORE_KEYS)
        base_info["title"] = dir_
        return render_template("overview.html", **base_info)
    except Exception as e:
        print(e.__repr__())
        raise e
        abort(500)


def experiment(base_dir):
    experiment_paths = request.args.getlist('exp')
    name = request.args.get('name', "")
    do_save = request.args.get('save', "")
    combi = request.args.get('combi', 'false')

    experiments = []

    if combi == "true":
        combi_exp = CombiExperimentReader(base_dir, experiment_paths, name=name)
        if do_save == "true":
            combi_exp.save()
        experiments = [combi_exp]
    else:
        # Get all Experiments
        for experiment_path in sorted(experiment_paths):
            exp = ExperimentReader(base_dir, experiment_path)
            experiments.append(exp)

    # Assign unique names
    exp_names = [exp.exp_name for exp in experiments]
    if len(exp_names) > len(set(exp_names)):
        for i, exp in enumerate(experiments):
            exp.exp_name += str(i)
    exp_names = [exp.exp_name for exp in experiments]

    # Site Content
    content = {}

    # Get config
    default_val = "-"
    combi_config = {}
    exp_configs = [exp.config.flat(False) for exp in experiments]
    diff_config_keys = list(Config.difference_config_static(*exp_configs).keys())
    config_keys = set([k for c in exp_configs for k in c.keys()])
    for k in sorted(config_keys):
        combi_config[k] = []
        for conf in exp_configs:
            combi_config[k].append(conf.get(k, default_val))
    config_keys = list(sorted(list(config_keys)))

    # Get results
    default_val = "-"
    combi_results = {}
    exp_results = [exp.get_results() for exp in experiments]
    result_keys = set([k for r in exp_results for k in r.keys()])
    for k in sorted(result_keys):
        combi_results[k] = []
        for res in exp_results:
            combi_results[k].append(res.get(k, default_val))
    result_keys = list(sorted(list(result_keys)))

    # Get images
    images = OrderedDict({})
    image_keys = set()
    image_path = {}
    for exp in experiments:
        exp_images = exp.get_images()
        img_groups = group_images(exp_images)
        images[exp.exp_name] = img_groups
        image_path[exp.exp_name] = exp.img_dir
        image_keys.update(list(img_groups.keys()))
    image_keys = list(image_keys)
    image_keys.sort()

    # Get Plots
    plots = OrderedDict({})
    for exp in experiments:
        exp_plots = exp.get_plots()
        plots[exp.exp_name] = exp_plots

    # Get logs
    logs_dict = OrderedDict({})
    for exp in experiments:
        exp_logs = [(os.path.basename(l), exp.exp_dir) for l in exp.get_logs()]
        logs_dict[exp.exp_name] = exp_logs

    content["title"] = experiments
    content["images"] = {"img_path": image_path, "imgs": images, "img_keys": image_keys}
    content["plots"] = {"plots": plots}
    content["config"] = {"exps": experiments, "configs": combi_config, "keys": config_keys,
                         "diff_keys": diff_config_keys}
    content["results"] = {"exps": exp_names, "results": combi_results, "keys": result_keys}
    content["logs"] = {"logs_dict": logs_dict}

    return render_template('experiment.html', **content)


def experiment_log(base_dir):
    experiment_path = request.args.get('exp')
    log_name = request.args.get('log')

    exp = ExperimentReader(base_dir, experiment_path)
    content = exp.get_log_file_content(log_name)

    print(experiment_path, log_name)

    return content


def experiment_remove(base_dir):
    experiment_paths = request.args.getlist('exp')

    # Get all Experiments
    for experiment_path in sorted(experiment_paths):
        exp = ExperimentReader(base_dir, experiment_path)
        exp.ignore_experiment()

    return ""


def experiment_plots(base_dir):
    experiment_paths = request.args.getlist('exp')
    combi = request.args.get('combi', 'false')
    experiments = []

    if combi == "true":
        combi_exp = CombiExperimentReader(base_dir, experiment_paths)
        # combi_exp.save()
        experiments = [combi_exp]
    else:
        # Get all Experiments
        for experiment_path in sorted(experiment_paths):
            exp = ExperimentReader(base_dir, experiment_path)
            experiments.append(exp)

    # Assign unique names
    exp_names = [exp.exp_name for exp in experiments]
    if len(exp_names) > len(set(exp_names)):
        for i, exp in enumerate(experiments):
            exp.exp_name += str(i)
    exp_names = [exp.exp_name for exp in experiments]

    results = []
    for exp in experiments:
        results.append(exp.get_results_log())
    results = merge_results(exp_names, results)

    graphs, traces = make_graphs(results, color_map=COLORMAP)
    graphs = [str(g) for g in graphs]

    return json.dumps({"graphs": graphs, "traces": traces})


def experiment_star(base_dir):
    experiment_path = request.args.get('exp')
    star_val = request.args.get('star')

    star_val = bool(int(star_val))

    print(experiment_path, star_val)

    exp = ExperimentReader(base_dir, experiment_path)
    exp.update_meta_info(star=star_val)

    return ""


def experiment_rename(base_dir):
    experiment_path = request.args.get('exp')
    new_name = request.args.get('name')

    print(experiment_path, new_name)

    exp = ExperimentReader(base_dir, experiment_path)
    exp.update_meta_info(name=new_name)

    return ""


def combine(base_dir):
    experiment_paths = request.args.getlist('exp')
    group_by_list = request.args.getlist('group')
    name = request.args.get('name', "")

    try:
        exps = []
        for ep in experiment_paths:
            expr = ExperimentReader(os.path.join(base_dir, ep))
            if expr.exp_info.get("state", "Combined") != "Combined":
                exps.append(expr)

        exp_groups = group_experiments_by(exps, group_by_list)

        combis = []
        for group in exp_groups:
            expc = CombiExperimentReader(base_dir="", exp_dirs=[e.work_dir for e in group], name=name)
            combis.append(expc)

        for expc in combis:
            expc.save()

        return "1"
    except Exception as e:
        return "0"

    return "0"


if __name__ == "__main__":
    start_browser()



from __future__ import print_function

import datetime
import json
import os
import re
import shutil
import warnings

try:
    import cPickle as pickle
except ImportError:
    import pickle

import numpy as np

from trixi.logger.abstractlogger import AbstractLogger
from trixi.logger.file.textfilelogger import TextFileLogger
from trixi.logger.file.numpyplotfilelogger import NumpyPlotFileLogger
from trixi.util import create_folder, MultiTypeEncoder, MultiTypeDecoder, Config


REPLACEMENTS = [("%Y", 4), ("%m", 2), ("%d", 2), ("%H", 2), ("%M", 2), ("%S", 2),
                ("%w", 1), ("%y", 2), ("%I", 2), ("%f", 6), ("%j", 3), ("%U", 2),
                ("%W", 2)]


class ExperimentLogger(AbstractLogger):
    """A single class for logging your experiments to file.

    It creates a experiment folder in your base folder and a folder structure to store your experiment files.
    The folder structure is::

        base_dir/
            new_experiment_folder/
                checkpoint/
                config/
                img/
                log/
                plot/
                result/
                save/


    """

    def __init__(self,
                 exp_name,
                 base_dir,
                 folder_format="%Y%m%d-%H%M%S_{experiment_name}",
                 resume=False,
                 text_logger_args=None,
                 plot_logger_args=None,
                 **kwargs):
        """
        Initializes the Experiment logger and creates the experiment folder structure

        Args:
            exp_name (str): The name of the experiment
            base_dir (str): The base directory in which the experiment folder will be created
            folder_format (str): The format for the naming of the experiment folder
            resume (bool): if True use the given folder and do not create new ones
            text_logger_args: Parameters for the TextFileLogger initialization
            plot_logger_args: Parameters for the NumpyPlotFileLogger initialization
        """

        super(ExperimentLogger, self).__init__(**kwargs)

        self.experiment_name = exp_name
        self.base_dir = base_dir
        self.folder_format = folder_format

        self.init_time = datetime.datetime.today()

        # try to make folder until successful or until counter runs out
        # this is necessary when multiple ExperimentLoggers start at the same time
        makedir_success = False
        makedir_counter = 100
        makedir_exception = None
        while (not makedir_success and makedir_counter > 0):
            try:
                self.folder_name = self.resolve_format(folder_format, resume)
                self.work_dir = os.path.join(base_dir, self.folder_name)
                if not resume:
                    create_folder(self.work_dir)
                makedir_success = True
            except FileExistsError as file_error:
                makedir_counter -= 1
            except Exception as e:
                makedir_exception = e
                makedir_counter -= 1
        if makedir_exception is not None:
            warnings.warn("Last exception encountered in makedir process:\n" +
                                 "{}\n".format(makedir_exception) +
                                 "There may or may not be a folder for the experiment to run in.", RuntimeWarning)

        self.config_dir = os.path.join(self.work_dir, "config")
        self.log_dir = os.path.join(self.work_dir, "log")
        self.checkpoint_dir = os.path.join(self.work_dir, "checkpoint")
        self.img_dir = os.path.join(self.work_dir, "img")
        self.plot_dir = os.path.join(self.work_dir, "plot")
        self.save_dir = os.path.join(self.work_dir, "save")
        self.result_dir = os.path.join(self.work_dir, "result")

        if not resume:
            create_folder(self.config_dir)
            create_folder(self.log_dir)
            create_folder(self.checkpoint_dir)
            create_folder(self.img_dir)
            create_folder(self.plot_dir)
            create_folder(self.save_dir)
            create_folder(self.result_dir)

        if text_logger_args is None:
            text_logger_args = {}
        if plot_logger_args is None:
            plot_logger_args = {}

        self.text_logger = TextFileLogger(self.log_dir, **text_logger_args)
        self.plot_logger = NumpyPlotFileLogger(
            self.img_dir, self.plot_dir, **plot_logger_args)

    def show_image(self, image, name, file_format=".png", **kwargs):
        """
        This function saves an image in the experiment img folder.

        Args:
            image(np.ndarray): image to be shown
            name(str): image title
            file_format (str): file format of the image
        """
        self.plot_logger.show_image(image, name, file_format=".png", **kwargs)

    def show_barplot(self, array, name, file_format=".png", **kwargs):
        """
        This function saves a barplot in the experiment plot folder.

        Args:
            array(np.ndarray): array to be plotted
            name(str): image title
            file_format (str): file format of the image
        """
        self.plot_logger.show_barplot(
            array, name, file_format=".png", **kwargs)

    def show_lineplot(self, y_vals, x_vals=None, name="lineplot", file_format=".png", **kwargs):
        """
        This function saves a line plot in the experiment plot folder.

        Args:
            x_vals: x values of the line
            y_vals: y values of the line
            name(str): image title
            file_format (str): file format of the image
        """
        self.plot_logger.show_lineplot(
            y_vals, x_vals, name, file_format=".png", **kwargs)

    def show_piechart(self, array, name, file_format=".png", **kwargs):
        """
        This function saves a piechart in the experiment plot folder.

        Args:
            array(np.ndarray): array to be plotted
            name(str): image title
            file_format (str): file format of the image
        """
        self.plot_logger.show_piechart(
            array, name, file_format=".png", **kwargs)

    def show_scatterplot(self, array, name, file_format=".png", **kwargs):
        """
        This function saves a scatterplot in the experiment plot folder.

        Args:
            array(np.ndarray): array to be plotted
            name(str): image title
            file_format (str): file format of the image
        """
        self.plot_logger.show_scatterplot(
            array, name, file_format=".png", **kwargs)

    def show_value(self, value, name=None, counter=None, tag=None, file_format=".png", **kwargs):
        """
        This function saves a value as a consequtive line plot.

        Args:
            value(np.ndarray): value to be plotted
            name(str): image title
            counter: y-value of the image (if not suppled simply increases for each call)
            tag: group/label for the value. Values with the same tag will be plotted in the same plot
            file_format (str): file format of the image
        """
        self.plot_logger.show_value(value, name, counter, tag, file_format, **kwargs)

    def show_text(self, text, name=None, logger="default", **kwargs):
        """
        Logs a text to a log file.

        Args:
            text: The text to be logged
            name: Name of the text
            logger: log file (in the experiment log folder) in which the text will be logged.
            **kwargs:

        """
        self.text_logger.show_text(text, name, logger, **kwargs)

    def show_boxplot(self, array, name, file_format=".png", **kwargs):
        """
        This function saves a boxplot in the experiment plot folder.

        Args:
            array(np.ndarray): array to be plotted
            name(str): image title
            file_format (str): file format of the image
        """
        self.plot_logger.show_boxplot(
            array, name, file_format=".png", **kwargs)

    def save_model(self):
        raise NotImplementedError

    def load_model(self):
        raise NotImplementedError

    def save_config(self, data, name, **kwargs):
        """
        Saves a config as a json file in the experiment config dir

        Args:
            data: The data to be stored as config json
            name: The name of the json file in which the data will be stored

        """

        if not name.endswith(".json"):
            name += ".json"
        data.dump(os.path.join(self.config_dir, name), **kwargs)

    def load_config(self, name, **kwargs):
        """
        Loads a config from a json file from the experiment config dir

        Args:
            name: the name of the config file

        Returns: A Config/ dict filled with the json file content

        """

        if not name.endswith(".json"):
            name += ".json"
        c = Config()
        c.load(os.path.join(self.config_dir, name), **kwargs)
        return c

    def save_checkpoint(self):
        raise NotImplementedError

    def load_checkpoint(self):
        raise NotImplementedError

    def save_result(self, data, name, indent=4, separators=(",", ": "), encoder_cls=MultiTypeEncoder, **kwargs):
        """
        Saves data as a json file in the experiment result dir

        Args:
            data: The data to be stored as result json
            name: name of the result json file
            indent: Indent for the json file
            separators: Separators for the json file
            encoder_cls: Encoder Class for the encoding to json

        """

        if not name.endswith(".json"):
            name += ".json"
        name = os.path.join(self.result_dir, name)
        create_folder(os.path.dirname(name))
        with open(name, "w") as jf:
            json.dump(data, jf,
                      cls=encoder_cls,
                      indent=indent,
                      separators=separators,
                      **kwargs)

    def save_dict(self, data, path, indent=4, separators=(",", ": "), encoder_cls=MultiTypeEncoder, **kwargs):
        """
        Saves a dict as a json file in the experiment save dir

        Args:
            data: The data to be stored as save file
            path: sub path in the save folder (or simply filename)
            indent: Indent for the json file
            separators: Separators for the json file
            encoder_cls: Encoder Class for the encoding to json
        """

        if not path.endswith(".json"):
            path += ".json"
        path = os.path.join(self.save_dir, path)
        create_folder(os.path.dirname(path))
        with open(path, "w") as jf:
            json.dump(data, jf,
                      cls=encoder_cls,
                      indent=indent,
                      separators=separators,
                      **kwargs)

    def load_dict(self, path):
        """
        Loads a json file as dict from a sub path in the experiment save dir

        Args:
            path: sub path to the file (starting from the experiment save dir)

        Returns: The restored data as a dict
        """

        if not path.endswith(".json"):
            path += ".json"
        path = os.path.join(self.save_dir, path)
        ret_val = dict()
        with open(path, "r") as df:
            ret_val = json.load(df, cls=MultiTypeDecoder)
        return ret_val

    def save_numpy_data(self, data, path):
        """
            Saves a numpy array in the experiment save dir

            Args:
                data: The array to be stored as a save file
                path: sub path in the save folder (or simply filename)
        """

        if not path.endswith(".npy"):
            path += ".npy"
        path = os.path.join(self.save_dir, path)
        create_folder(os.path.dirname(path))
        np.save(path, data)

    def load_numpy_data(self, path):
        """
        Loads a numpy file from a sub path in the experiment save dir

        Args:
            path: sub path to the file (starting from the experiment save dir)

        Returns: The restored numpy array
        """

        if not path.endswith(".npy"):
            path += ".npy"
        path = os.path.join(self.save_dir, path)
        return np.load(path)

    def save_pickle(self, data, path):
        """
            Saves a object data in the experiment save dir via pickle

            Args:
                data: The data to be stored as a save file
                path: sub path in the save folder (or simply filename)
        """

        path = os.path.join(self.save_dir, path)
        create_folder(os.path.dirname(path))
        with open(path, "wb") as out:
            pickle.dump(data, out)

    def load_pickle(self, path):
        """
        Loads a object via pickle from a sub path in the experiment save dir

        Args:
            path: sub path to the file (starting from the experiment save dir)

        Returns: The restored object
        """

        path = os.path.join(self.save_dir, path)
        with open(path, "rb") as in_:
            return pickle.load(in_)

    def save_file(self, filepath, path=None):
        """
        Copies a file to the experiment save dir

        Args:
            filepath: Path to the file to be copied to the experiment save dir
            path: sub path to the target file (starting from the experiment save dir, does not have to exist yet)

        """

        if path is None:
            target_dir = self.save_dir
        else:
            target_dir = os.path.join(self.save_dir, path)
        filename = os.path.basename(filepath)
        shutil.copy(filepath, os.path.join(target_dir, filename))

    def resolve_format(self, input_, resume):
        """
        Given some input pattern, tries to find the best matching folder name by resolving the format. Options are:
         - Run-number: {run_number}
         - Time: "%Y%m%d-%H%M%S
         - Member variables (e.g experiment_name) : {variable_name} (e.g. {experiment_name})


        Args:
            input_: The format to be resolved
            resume: Flag if folder should be resumed

        Returns: The resolved folder name

        """

        if resume:

            pattern = input_[:]

            for find in re.findall("{[\w\:]+}", pattern):
                run_match = re.search("(?<=\{run_number\:0)\d+(?=d\})", find)
                if find == "{run_number}":
                    pattern = pattern.replace("{run_number}", "\d+")
                elif run_match:
                    length = int(run_match.group(0))
                    pattern = re.sub("\{run_number\:\d+d\}",
                                     "\\d{" + str(length) + "}", pattern)
                else:
                    if find[1:-1] in self.__dict__:
                        pattern = pattern.replace(
                            find, self.__dict__[find[1:-1]])

            for r in REPLACEMENTS:
                pattern = pattern.replace(r[0], "\\d{" + str(r[1]) + "}")

            return list(filter(lambda x: re.match(pattern, x),
                               sorted(os.listdir(self.base_dir))))[-1]

        if "%" in input_:
            input_ = self.init_time.strftime(input_)

        if "{" not in input_:
            return input_

        run_number = 1
        while os.path.exists(os.path.join(self.base_dir,
                                          input_.format(run_number=run_number,
                                                        **self.__dict__))):
            # if the folder already exists, for example if two jobs are launched
            # at the same time, append run_number
            if "{run_number}" not in input_:
                input_ = input_ + "_{run_number}"

            run_number += 1

        return input_.format(run_number=run_number, **self.__dict__)



from __future__ import division, print_function

import atexit
from collections import defaultdict
import inspect

import os

if os.name == "nt":
    IS_WINDOWS = True
    from queue import Queue
    from threading import Thread as Process
else:
    IS_WINDOWS = False
    try:
        from torch.multiprocessing import Queue, Process

        print("Using torch multi processing")
    except ImportError:
        from multiprocessing import Queue, Process
import sys
import traceback

import numpy as np

from trixi.logger.abstractlogger import AbstractLogger, convert_params
from trixi.util import ExtraVisdom


class NumpyVisdomLogger(AbstractLogger):
    """
    Visual logger, inherits the AbstractLogger and plots/ logs numpy arrays/ values on a Visdom server.
    """

    def __init__(self, exp_name="main", server="http://localhost", port=8080, auto_close=True, auto_start=False,
                 auto_start_ports=(8080, 8000), **kwargs):
        """
        Creates a new NumpyVisdomLogger object.

        Args:
            name: The name of the visdom environment
            server: The address of the visdom server
            port: The port of the visdom server
            auto_close: Close all objects and kill process at the end of the python script
            auto_start: Flag, if it should try to start a visdom server on the given ports
            auto_start_ports: Ordered list of ports, to try to start a visdom server on (only on the first available
            port)
        """
        super(NumpyVisdomLogger, self).__init__(**kwargs)

        if auto_start:
            auto_port = start_visdom(auto_start_ports)
            if auto_port != -1:
                port = auto_port
                server = "http://localhost"

        self.name = exp_name
        self.server = server
        self.port = port

        self.vis = ExtraVisdom(env=self.name, server=self.server, port=self.port)

        self._value_counter = defaultdict(dict)
        self._3d_histograms = dict()

        self._queue = Queue()
        self._process = Process(target=self.__show, args=(self._queue,))

        if auto_close:
            # atexit.register(self.close_all)
            if not IS_WINDOWS:
                atexit.register(self.exit)
            atexit.register(self.save_vis)

        self._process.start()

    def __show(self, queue):
        """
        Loop for the internal process to process all visualization tasks

        Args:
            queue: queue with all visualization tasks
        """

        while True:

            vis_task = queue.get()

            try:

                show_fn = self.show_funcs[vis_task["type"]]
                show_fn(self, **vis_task)

            except Exception as e:

                error = sys.exc_info()[0]
                msg = traceback.format_exc()
                print("Error {}: {}".format(error, msg))

    @convert_params
    def show_image(self, image, name=None, title=None, caption=None, env_appendix="", opts=None, **kwargs):
        """
        Displays an image in a window/pane at the visdom server

        Args:
            image: The image array to be displayed
            name: The name of the image window
            title: The title of the image window
            caption: The of the image, displayed in the window
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "image",
            "image": image,
            "name": name,
            "title": title,
            "caption": caption,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_image(self, image, name=None, title=None, caption=None, env_appendix="", opts=None, **kwargs):
        """
        Internal show_image method, called by the internal process.
        This function does all the magic.
        """

        if opts is None:
            opts = {}
        opts = opts.copy()
        opts.update(dict(
            title=title,
            caption=caption
        ))

        win = self.vis.image(
            img=image,
            win=name,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    @convert_params
    def show_images(self, images, name=None, title=None, caption=None, env_appendix="", opts=None, **kwargs):
        """
        Displays multiple images in a window/pane at a visdom server

        Args:
            images: The image array to be displayed
            name: The name of the window
            title: The title of the image window
            caption: The of the image, displayed in the window
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "images",
            "images": images,
            "name": name,
            "title": title,
            "caption": caption,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_images(self, images, name=None, title=None, caption=None, env_appendix="", opts=None, **kwargs):
        """
       Internal show_images method, called by the internal process.
       This function does all the magic.
        """
        if opts is None:
            opts = {}
        else:
            if 'nrow' in opts.keys():
                nrow = opts['nrow']
            else:
                nrow = 8  # as defined in function
        opts = opts.copy()
        opts.update(dict(
            title=title,
            caption=caption
        ))

        win = self.vis.images(
            tensor=images,
            win=name,
            env=self.name + env_appendix,
            opts=opts,
            # nrow=nrow
        )

        return win

    @convert_params
    def show_value(self, value, name=None, counter=None, tag=None, show_legend=True, env_appendix="", opts=None,
                   **kwargs):
        """
        Creates a line plot that is automatically appended with new values and plots it to a visdom server.

        Args:
            value: Value to be plotted / appended to the graph (y-axis value)
            name: The name of the window
            counter: counter, which tells the number of the sample (with the same name) (x-axis value)
            tag: Tag, grouping similar values. Values with the same tag will be plotted in the same plot
            show_legend (bool): Flag, if it should display a legend
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "value",
            "value": value,
            "name": name,
            "counter": counter,
            "tag": tag,
            "show_legend": show_legend,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_value(self, value, name=None, counter=None, tag=None, show_legend=True, env_appendix="", opts=None,
                     **kwargs):
        """
       Internal show_value method, called by the internal process.
       This function does all the magic.
        """

        if opts is None:
            opts = {}
        value_dim = 1
        if isinstance(value, (list, tuple, np.ndarray)):
            value_dim = len(value)
            value = np.asarray([value])
        else:
            value = np.asarray([value])
        if tag is None:
            tag = ""
            if name is not None:
                tag = name
        if "showlegend" not in opts and show_legend:
            opts["showlegend"] = True

        up_str = None
        if tag is not None and tag in self._value_counter:
            up_str = "append"

        if tag is not None and tag in self._value_counter and name in self._value_counter[tag]:
            if counter is None:
                self._value_counter[tag][name] += 1
            else:
                self._value_counter[tag][name] += counter - self._value_counter[tag][name]
        else:
            if counter is None:
                counter = 0
            if value_dim == 1:
                self._value_counter[tag][name] = np.array([counter])
            else:
                self._value_counter[tag][name] = np.array([[counter] * value_dim])

        opts = opts.copy()
        opts.update(dict(
            title=tag,
        ))

        display_name = tag
        if value_dim != 1:
            display_name = None

        win = self.vis.line(
            Y=value,
            X=self._value_counter[tag][name],
            win=tag,
            name=name,
            update=up_str,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    @convert_params
    def show_text(self, text, name=None, env_appendix="", opts=None, **kwargs):
        """
        Displays a text in a visdom window

        Args:
             text: The text to be displayed
             name: The name of the window
             env_appendix: appendix to the environment name, if used the new env is env+env_appendix
             opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "text",
            "text": text,
            "name": name,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_text(self, text, name=None, env_appendix="", opts=None, **kwargs):
        """
       Internal show_text method, called by the internal process.
       This function does all the magic.
        """

        text = text.replace("\n", "<br>")

        if opts is None:
            opts = {}
        win = self.vis.text(
            text=text,
            win=name,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    @convert_params
    def show_progress(self, num, total=None, name=None, env_appendix="", opts=None, **kwargs):
        """
        Shows the progress as a pie chart.

        Args:
            num: Current progress. Either a relative value (0 <= num <= 1) or a absolute value if total is given (
            but has to be smaller than total)
            total: This if the total number of iterations.
            name: The name of the window
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "progress",
            "num": num,
            "total": total,
            "name": name,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_progress(self, num, total=None, name=None, env_appendix="", opts=None, **kwargs):
        """
       Internal show_progress method, called by the internal process.
       This function does all the magic.
        """

        if opts is None:
            opts = {}
        if total is None:
            if 0 <= num <= 1:
                total = 1000.
                num = 1000. * num
            else:
                raise AttributeError("Either num has to be a ratio between 0 and 1 or give a valid total number")
        else:
            if num > total:
                raise AttributeError("Num has to be smaller than total")

        if name is None:
            name = "progress"

        x = np.asarray([num, total - num])

        opts = opts.copy()
        opts.update(dict(
            legend=['Done', 'To-Do'],
            title="Progress")
        )

        win = self.vis.pie(
            X=x,
            win=name,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    @convert_params
    def show_histogram(self, array, name=None, bins=30, env_appendix="", opts=None, **kwargs):
        """
        Displays the histogramm of an array.

        Args:
            array: The array the histogram is calculated of
            name: The name of the window
            bins: Number of bins (== bars) in the histogram
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "histogram",
            "array": array,
            "bins": bins,
            "name": name,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_histogram(self, array, name=None, bins=30, env_appendix="", opts=None, **kwargs):
        """
       Internal show_histogram method, called by the internal process.
       This function does all the magic.
        """
        if opts is None:
            opts = {}
        opts = opts.copy()
        opts.update(dict(
            title=name,
            numbins=bins
        ))

        win = self.vis.histogram(
            X=array,
            win=name,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    @convert_params
    def show_histogram_3d(self, array, name, bins=50, env_appendix="", opts=None, **kwargs):
        """
        Displays a history of histograms as consequtive lines in a 3d space (similar to tensorflow)

        Args:
            array: New sample of the array that should be plotted (i.e. results in one line of the 3d histogramm)
            name: The name of the window
            bins: Number of bins in the histogram
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "histogram_3d",
            "array": array,
            "bins": bins,
            "name": name,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_histogram_3d(self, array, name, bins=50, env_appendix="", opts=None, **kwargs):
        """
       Internal show_histogram_3d method, called by the internal process.
       This function does all the magic.
        """

        if opts is None:
            opts = {}
        if name not in self._3d_histograms:
            self._3d_histograms[name] = []

        self._3d_histograms[name].append(array)

        if len(self._3d_histograms[name]) > 50:

            n_histo = len(self._3d_histograms[name]) - 20
            every_n = n_histo // 30 + 1

            x = self._3d_histograms[name][:-20][0::every_n] + self._3d_histograms[name][-20:]
        else:
            x = self._3d_histograms[name]

        opts = opts.copy()
        opts.update(dict(
            title=name,
            numbins=bins
        ))

        win = self.vis.histogram_3d(
            X=x,
            win=name,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    @convert_params
    def show_barplot(self, array, legend=None, rownames=None, name=None, env_appendix="", opts=None, **kwargs):
        """
        Displays a bar plot from an array

        Args:
            array: array of shape NxM where N is the nomber of rows and M is the number of elements in the row.
            legend: list of legend names. Has to have M elements.
            rownames: list of row names. Has to have N elements.
            name: The name of the window
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "barplot",
            "array": array,
            "legend": legend,
            "rownames": rownames,
            "name": name,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_barplot(self, array, legend=None, rownames=None, name=None, env_appendix="", opts=None, **kwargs):
        """
       Internal show_barplot method, called by the internal process.
       This function does all the magic.
        """
        if opts is None:
            opts = {}
        opts = opts.copy()
        opts.update(dict(
            stacked=False,
            legend=legend,
            rownames=rownames,
            title=name
        ))

        win = self.vis.bar(
            X=array,
            win=name,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    @convert_params
    def show_lineplot(self, y_vals, x_vals=None, name=None, env_appendix="", show_legend=True, opts=None, **kwargs):
        """
        Displays (multiple) lines plot, given values Y (and optional the corresponding X values)

        Args:
            y_vals: Array of shape MxN , where M is the number of points and N is the number of different line
            x_vals: Has to have the same shape as Y: MxN. For each point in Y it gives the corresponding X value (if
            not set the points are assumed to be equally distributed in the interval [0, 1] )
            name: The name of the window
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "lineplot",
            "x_vals": x_vals,
            "y_vals": y_vals,
            "name": name,
            "show_legend": show_legend,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_lineplot(self, y_vals, x_vals=None, name=None, env_appendix="", show_legend=True, opts=None, **kwargs):
        """
       Internal show_lineplot method, called by the internal process.
       This function does all the magic.
        """

        if opts is None:
            opts = {}
        opts = opts.copy()
        opts.update(dict(
            title=name,
        ))

        if "showlegend" not in opts and show_legend:
            opts["showlegend"] = True

        win = self.vis.line(
            X=x_vals,
            Y=y_vals,
            win=name,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    @convert_params
    def show_boxplot(self, x_vals, name=None, env_appendix="", opts=None, **kwargs):
        """
        Displays a box plot, given values X

        Args:
            x_vals: Array of shape MxN , where M is the number of points and N are the number of groups
            name: The name of the window
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "boxplot",
            "x_vals": x_vals,
            "name": name,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_boxplot(self, x_vals, name=None, env_appendix="", opts=None, **kwargs):
        """
       Internal show_lineplot method, called by the internal process.
       This function does all the magic.
        """

        if opts is None:
            opts = {}
        opts = opts.copy()
        opts.update(dict(
            title=name,
        ))

        win = self.vis.boxplot(
            X=x_vals,
            win=name,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    @convert_params
    def show_surfaceplot(self, x_vals, name=None, env_appendix="", opts=None, **kwargs):
        """
        Displays a surface plot given values X

        Args:
            x_vals: Array of shape MxN
            name: The name of the window
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "surfaceplot",
            "x_vals": x_vals,
            "name": name,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_surfaceplot(self, x_vals, name=None, env_appendix="", opts=None, **kwargs):
        """
       Internal show_surfaceplot method, called by the internal process.
       This function does all the magic.
        """

        if opts is None:
            opts = {}
            opts.update(dict(
                title=name,
                colormap='Viridis',
                xlabel='X',
                ylabel='Y'
            ))
        opts = opts.copy()

        win = self.vis.surf(X=x_vals,
                            win=name,
                            env=self.name + env_appendix,
                            opts=opts, )

        return win

    @convert_params
    def show_contourplot(self, x_vals, name=None, env_appendix="", opts=None, **kwargs):
        """
        Displays a contour plot

        Args:
            x_vals: Array of shape MxN
            name: The name of the window
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "contourplot",
            "x_vals": x_vals,
            "name": name,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_contourplot(self, x_vals, name=None, env_appendix="", opts=None, **kwargs):
        """
       Internal show_contourplot method, called by the internal process.
       This function does all the magic.
        """

        if opts is None:
            opts = {}
            opts.update(dict(
                title=name,
                colormap='Viridis',
                xlabel='feature',
                ylabel='batch'
            ))
        opts = opts.copy()

        win = self.vis.contour(X=x_vals,
                               win=name,
                               env=self.name + env_appendix,
                               opts=opts, )

        return win

    @convert_params
    def show_scatterplot(self, array, labels=None, name=None, env_appendix="", opts=None, **kwargs):
        """
        Displays a scatter plots, with the points given in array

        Args:
            array: A 2d array with size N x dim, where each element i \in N at X[i] results in a a 2d (if dim = 2)/
            3d (if dim = 3) point.
            labels: For each point a int label (starting from 1) can be given. Has to be an array of size N.
            name: The name of the window
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "scatterplot",
            "array": array,
            "labels": labels,
            "name": name,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_scatterplot(self, array, labels=None, name=None, env_appendix="", opts=None, **kwargs):
        """
       Internal show_scatterplot method, called by the internal process.
       This function does all the magic.
        """

        if opts is None:
            opts = {}
        opts.update(dict(
            title=name,
        ))

        win = self.vis.scatter(
            X=array,
            Y=labels,
            win=name,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    @convert_params
    def show_piechart(self, array, name=None, env_appendix="", opts=None, **kwargs):
        """
        Displays a pie chart.

        Args:
            array: Array of positive integers. Each integer will be presented as a part of the pie (with the total
            as the sum of all integers)
            name: The name of the window
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "piechart",
            "array": array,
            "name": name,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_piechart(self, array, name=None, env_appendix="", opts=None, **kwargs):
        """
       Internal show_piechart method, called by the internal process.
       This function does all the magic.
        """

        if opts is None:
            opts = {}
        opts.update(dict(
            title=name
        ))

        win = self.vis.pie(
            X=array,
            win=name,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    @convert_params
    def show_svg(self, svg, name=None, env_appendix="", opts=None, **kwargs):
        """
        Displays a svg file.

        Args:
            svg: Filename of the svg file which should be displayed
            name: The name of the window
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "svg",
            "svg": svg,
            "name": name,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_svg(self, svg, name=None, env_appendix="", opts=None, **kwargs):
        """
       Internal show_svg method, called by the internal process.
       This function does all the magic.
        """

        if opts is None:
            opts = {}
        with open(svg, 'r') as fileobj:
            svgstr = fileobj.read()

        opts.update(dict(
            title=name
        ))

        win = self.vis.svg(
            svgstr=svgstr,
            win=name,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    def show_values(self, val_dict):
        """A util function for multiple values. Simple plots all values in a dict, there the window name is the key
        in the dict and the plotted value is the value of a dict (Simply calls the show_value function).

        Args:
            val_dict: Dict with key, values pairs which will be plotted
        """

        for name, value in val_dict:
            self.show_value(value, name)

    @convert_params
    def add_to_graph(self, y_vals, x_vals=None, name=None, legend_name=None, append=True, env_appendix="", opts=None,
                     **kwargs):
        """
        Displays (multiple) lines plot, given values Y (and optional the corresponding X values)

        Args:
            y_vals: Array of shape MxN , where M is the number of points and N is the number of different line
            x_vals: Has to have the same shape as Y: MxN. For each point in Y it gives the corresponding X value (if
            not set the points are assumed to be equally distributed in the interval [0, 1] )
            name: The name of the window
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "add",
            "x_vals": x_vals,
            "y_vals": y_vals,
            "name": name,
            "legend_name": legend_name,
            "append": append,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __add_to_graph(self, y_vals, x_vals=None, name=None, legend_name=None, append=True, env_appendix="", opts=None,
                       **kwargs):
        """
       Internal show_lineplot method, called by the internal process.
       This function does all the magic.
        """

        if opts is None:
            opts = {}
        opts.update(dict(
            title=name,
        ))

        win = self.vis.updateTrace(
            X=x_vals,
            Y=y_vals,
            win=name,
            name=legend_name,
            append=append,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    @convert_params
    def show_matplot_plt(self, plt, name=None, title=None, caption=None, env_appendix="", opts=None, **kwargs):
        """
        Displays an matplotlib figure in a window/pane at the visdom server

        Args:
            plt: The matplotlib figure/plot to be displayed
            name: The name of the image window
            title: The title of the image window
            caption: The of the image, displayed in the window
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        if opts is None:
            opts = {}
        vis_task = {
            "type": "matplot_plt",
            "plt": plt,
            "name": name,
            "title": title,
            "caption": caption,
            "env_appendix": env_appendix,
            "opts": opts
        }
        self._queue.put_nowait(vis_task)

    def __show_matplot_plt(self, plt, name=None, title=None, caption=None, env_appendix="", opts=None, **kwargs):
        """
        Internal show_plt_figure method, called by the internal process.
        This function does all the magic.
        """

        if opts is None:
            opts = {}
        opts = opts.copy()
        opts.update(dict(
            title=title,
            caption=caption
        ))

        win = self.vis.matplot(
            plot=plt,
            win=name,
            env=self.name + env_appendix,
            opts=opts
        )

        return win

    @convert_params
    def show_plotly_plt(self, figure, name=None, env_appendix="", **kwargs):
        """
        Displays an plotly figure in a window/pane at the visdom server

        Args:
            figure: The plotly figure/plot to be displayed
            name: The name of the image window
            title: The title of the image window
            caption: The of the image, displayed in the window
            env_appendix: appendix to the environment name, if used the new env is env+env_appendix
            opts: opts dict for the ploty/ visdom plot, i.e. can set window size, en/disable ticks,...
        """

        vis_task = {
            "type": "plotly_plt",
            "figure": figure,
            "name": name,
            "env_appendix": env_appendix,
        }
        self._queue.put_nowait(vis_task)

    def __show_plotly_plt(self, figure, name=None, env_appendix="", **kwargs):
        """
        Internal show_plt_figure method, called by the internal process.
        This function does all the magic.
        """

        win = self.vis.plotlyplot(
            figure=figure,
            win=name,
            env=self.name + env_appendix,
        )

        return win

    @convert_params
    def send_data(self, data, name=None, layout=None, endpoint='events', append=False, **kwargs):
        """
        Sends data to a visdom server.

        Args:
            data: The data to be send (has to be in the plotly / visdom format, see the offical visdom github page)
            name: The name of the window
            layout: Layout of the data
            endpoint: Endpoint to recieve the data (or at least to which visdom endpoint to send it to)
            append: Flag, if data should be appended
        """

        if layout is None:
            layout = {}
        vis_task = {
            "type": "data",
            "data": data,
            "layout": layout,
            "name": name,
            "endpoint": endpoint,
            "append": append
        }
        self._queue.put_nowait(vis_task)

    def __send_data(self, data, name=None, layout=None, endpoint='events', append=False, **kwargs):
        """
       Internal show_lineplot method, called by the internal process.
       This function does all the magic.
        """

        if layout is None:
            layout = {}

        if not isinstance(data, (list, type)):
            data = [data]

        win = self.vis._send({'data': data,
                              'layout': layout,
                              'win': name,
                              'append': append}, endpoint=endpoint)

        return win

    def close_all(self):
        """Closes all visdom windows."""
        self.vis.close()

    def save_vis(self):
        self.vis.save([self.name])

    def exit(self):
        """Kills the internal process."""
        if self._process is not None:
            self._process.terminate()

    show_funcs = {
        "image": __show_image,
        "images": __show_images,
        "value": __show_value,
        "text": __show_text,
        "progress": __show_progress,
        "histogram": __show_histogram,
        "histogram_3d": __show_histogram_3d,
        "barplot": __show_barplot,
        "boxplot": __show_boxplot,
        "surfaceplot": __show_surfaceplot,
        "contourplot": __show_contourplot,
        "lineplot": __show_lineplot,
        "scatterplot": __show_scatterplot,
        "piechart": __show_piechart,
        "svg": __show_svg,
        "add": __add_to_graph,
        "data": __send_data,
        "matplot_plt": __show_matplot_plt,
        "plotly_plt": __show_plotly_plt,
    }


def start_visdom(port_list=(8080, 8000)):
    """
    Starts a visdom server on a given port

    Args:
        port_list: Priority list of port. Will start a visdom server on the first available port.

    """
    import time
    from multiprocessing import Process
    import visdom.server

    from trixi.util import PyLock

    lock_id = "visdom_lock"

    def is_port_available(port, verbose=False):

        import socket
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.bind(("127.0.0.1", port))
        except socket.error as e:
            if e.errno == 98:
                if verbose:
                    print("Port {} is already in use".format(port))
            else:
                if verbose:
                    print(e)
            s.close()
            return False
        s.close()
        return True

    def _start_visdom(port):
        p = Process(target=visdom.server.start_server, kwargs={"port": port, "base_url": ''})
        p.start()
        atexit.register(p.terminate)
        time.sleep(20)
        return True

    lock = PyLock(lock_id, timeout=10)

    i = 0
    while i < len(port_list):
        port = port_list[i]

        try:
            lock.__enter__()
            if is_port_available(port):
                if _start_visdom(port):
                    lock.__exit__(0, 0, 0)
                    print("Started Visdom on Port:", port)
                    return port
            else:
                i += 1
        except Exception:
            pass

    return -1



import os
from collections import defaultdict
import re

import colorlover as cl
import numpy as np
import plotly.graph_objs as go
from flask import Markup
from plotly.offline import plot
from scipy.signal import savgol_filter

from trixi.experiment_browser.experimentreader import ExperimentReader

# These keys will be ignored when in a config file
from trixi.util import Config

IGNORE_KEYS = ("name",
               "experiment_dir",
               "work_dir",
               "config_dir",
               "log_dir",
               "checkpoint_dir",
               "img_dir",
               "plot_dir",
               "save_dir",
               "result_dir",
               "time",
               "state")

# Set the color palette for plots
COLORMAP = cl.scales["8"]["qual"]["Dark2"]


def process_base_dir(base_dir, view_dir="", default_val="-", short_len=25, ignore_keys=IGNORE_KEYS):
    """Create an overview table of all experiments in the given directory.

    Args:
        directory (str): A directory containing experiment folders.
        default_val (str): Default value if an entry is missing.
        short_len (int): Cut strings to this length. Full string in alt-text.

    Returns:
        dict: {"ccols": Columns for config entries,
               "rcols": Columns for result entries,
               "rows": The actual data}

    """

    full_dir = os.path.join(base_dir, view_dir)

    config_keys = set()
    result_keys = set()
    exps = []
    non_exps = []

    ### Load Experiments with keys / different param values
    for sub_dir in sorted(os.listdir(full_dir)):
        dir_path = os.path.join(full_dir, sub_dir)
        if os.path.isdir(dir_path):
            try:
                exp = ExperimentReader(full_dir, sub_dir)
                if exp.ignore:
                    continue
                config_keys.update(list(exp.config.flat().keys()))
                result_keys.update(list(exp.get_results().keys()))
                exps.append(exp)
            except Exception as e:
                print("Could not load experiment: ", dir_path)
                print(e)
                print("-" * 20)
                non_exps.append(os.path.join(view_dir, sub_dir))

    ### Get not common val keys
    diff_keys = list(Config.difference_config_static(*[xp.config.flat() for xp in exps]))

    ### Remove unwanted keys
    config_keys -= set(ignore_keys)
    result_keys -= set(ignore_keys)

    ### Generate table rows
    sorted_c_keys1 = sorted([c for c in config_keys if c in diff_keys], key=lambda x: str(x).lower())
    sorted_c_keys2 = sorted([c for c in config_keys if c not in diff_keys], key=lambda x: str(x).lower())
    sorted_r_keys = sorted(result_keys, key=lambda x: str(x).lower())

    rows = []
    for exp in exps:
        config_row = []
        for key in sorted_c_keys1:
            attr_strng = str(exp.config.flat().get(key, default_val))
            config_row.append((attr_strng, attr_strng[:short_len]))
        for key in sorted_c_keys2:
            attr_strng = str(exp.config.flat().get(key, default_val))
            config_row.append((attr_strng, attr_strng[:short_len]))
        result_row = []
        for key in sorted_r_keys:
            attr_strng = str(exp.get_results().get(key, default_val))
            result_row.append((attr_strng, attr_strng[:short_len]))

        name = exp.exp_name
        time = exp.exp_info.get("time", default_val) if "time" in exp.exp_info else exp.config.get("time", default_val)
        state = exp.exp_info.get("state", default_val) if "state" in exp.exp_info else exp.config.get("state",
                                                                                                      default_val)
        epoch = exp.exp_info.get("epoch", default_val) if "epoch" in exp.exp_info else exp.config.get("epoch",
                                                                                                      default_val)

        rows.append((os.path.relpath(exp.work_dir, base_dir),
                     exp.star,
                     str(name),
                     str(time),
                     str(state),
                     str(epoch),
                     config_row, result_row))

    return {"ccols1": sorted_c_keys1, "ccols2": sorted_c_keys2, "rcols": sorted_r_keys, "rows": rows, "noexp": non_exps}


def group_images(images):
    images.sort()
    group_dict = defaultdict(list)

    for img in images:
        filename = img.split(os.sep + "img" + os.sep)[1]
        base_name = os.path.splitext(filename)[0]
        number_groups = re.findall("\d+\.\d+", base_name)
        if len(number_groups) == 0:
            base_name = ''.join(e for e in base_name if e.isalpha())
        else:
            base_name = base_name.replace(number_groups[0], "")

        group_dict[base_name].append(filename)

    return group_dict


def make_graphs(results, trace_options=None, layout_options=None, color_map=COLORMAP):
    """Create plot markups.

    This converts results into plotly plots in markup form. Results in a common
    group will be placed in the same plot.

    Args:
        results (dict): Dictionary

    """

    if trace_options is None:
        trace_options = {}
    if layout_options is None:
        layout_options = {
            "legend": dict(
                orientation="v",
                xanchor="left",
                x=0,
                yanchor="top",
                y=-0.1,
                font=dict(
                    size=8,
                )
            )
        }

    graphs = []
    trace_counters = []

    for group in sorted(results):

        layout = go.Layout(title=group, **layout_options)
        traces = []

        for r, result in enumerate(sorted(results[group])):

            y = np.array(results[group][result]["data"])
            x = np.array(results[group][result]["counter"])

            do_filter = len(y) >= 1000
            opacity = 0.2 if do_filter else 1.

            if "min" in results[group][result] and "max" in results[group][result]:
                min_ = np.array(results[group][result]["min"])
                max_ = np.array(results[group][result]["max"])
                fill_color = color_map[r % len(color_map)][:3] + "a" + color_map[r % len(color_map)][3:-1] + ",0.1)"
                upper_bound = go.Scatter(x=x, y=max_, name=result, legendgroup=result, showlegend=False,
                                         mode='lines', line=dict(width=0), hoverinfo='none',
                                         fillcolor=fill_color, **trace_options)
                lower_bound = go.Scatter(x=x, y=min_, name=result, legendgroup=result, showlegend=False,
                                         mode='lines', fill="tonexty", line=dict(width=0), hoverinfo='none',
                                         fillcolor=fill_color, **trace_options)
                traces.append(upper_bound)
                traces.append(lower_bound)
                traces.append(go.Scatter(x=x, y=y, opacity=opacity, name=result, legendgroup=result,
                                         line=dict(color=color_map[r % len(color_map)]), **trace_options))
            elif do_filter:
                def filter_(x):
                    return savgol_filter(x, max(5, 2 * (len(y) // 50) + 1), 3)

                traces.append(go.Scatter(x=x, y=y, opacity=opacity, name=result, legendgroup=result, showlegend=False,
                                         line=dict(color=color_map[r % len(color_map)]), **trace_options))
                traces.append(go.Scatter(x=x, y=filter_(y), name=result, legendgroup=result,
                                         line=dict(color=color_map[r % len(color_map)]), **trace_options))
            else:
                traces.append(go.Scatter(x=x, y=y, opacity=opacity, name=result, legendgroup=result,
                                         line=dict(color=color_map[r % len(color_map)]), **trace_options))

        trace_counters.append(len(results[group]))
        graphs.append(Markup(plot({"data": traces, "layout": layout},
                                  output_type="div",
                                  include_plotlyjs=False,
                                  show_link=False)))

    return graphs, trace_counters


def merge_results(experiment_names, result_list):
    merged_results = {}

    for r, result in enumerate(result_list):
        for label in result.keys():
            if label not in merged_results:
                merged_results[label] = {}
            for key in result[label].keys():
                new_key = "_".join([experiment_names[r], key])
                merged_results[label][new_key] = result[label][key]

    return merged_results



import numpy as np
import telegram
import torch
import torchvision
from trixi.util.util import get_image_as_buffered_file

from trixi.logger.plt.numpyseabornimageplotlogger import NumpySeabornImagePlotLogger


class TelegramMessageLogger(NumpySeabornImagePlotLogger):
    """
    Telegram logger, inherits the NumpySeabornImagePlotLogger and sends plots/logs to a chat via a Telegram bot.
    """

    def __init__(self, token, chat_id, exp_name=None, *args, **kwargs):
        """
        Creates a new TelegramMessageLogger object.

        Args:
            token (str): The token of the Telegram bot used.
            chat_id (str): The chat ID for the chat between the user and the Telegram bot.
        """
        super(TelegramMessageLogger, self).__init__(**kwargs)

        self.token = token
        self.chat_id = chat_id
        self.bot = telegram.Bot(token=self.token)
        self.exp_name = exp_name

    def process_params(self, f, *args, **kwargs):
        """
        Inherited "decorator": convert PyTorch variables and Tensors to numpy arrays
        """

        # convert args
        args = (a.detach().cpu().numpy() if torch.is_tensor(a) else a for a in args)

        # convert kwargs
        for key, data in kwargs.items():
            if torch.is_tensor(data):
                kwargs[key] = data.detach().cpu().numpy()

        return f(self, *args, **kwargs)

    def show_text(self, text, *args, **kwargs):
        """
        Sends a text to a chat using an existing Telegram bot.

        Args:
            text (str): Text message to be sent to the bot.
        """
        if self.exp_name is not None:
            text = self.exp_name + ":\n" + text
        try:
            self.bot.send_message(chat_id=self.chat_id, text=text)
        except Exception as e:
            print("Could not send text to telegram")

    def show_image(self, image, *args, **kwargs):
        """
        Sends an image file to a chat using an existing Telegram bot.

        Args:
            image (str or np array): Path to the image file to be sent to the chat.
        """
        try:
            if isinstance(image, str):
                with open(image, 'rb') as img_file:
                    self.bot.send_photo(chat_id=self.chat_id, photo=img_file)
            elif isinstance(image, np.ndarray):
                buf = get_image_as_buffered_file(image)

                self.bot.send_photo(chat_id=self.chat_id, photo=buf)

        except Exception as e:
            print("Could not send image to telegram")

    def show_image_grid(self, image_array, name=None, nrow=8, padding=2,
                        normalize=False, range=None, scale_each=False, pad_value=0, *args, **kwargs):
        """
        Sends an array of images to a chat using  an existing Telegram bot. (Requires torch and torchvision)


        Args:
            image_array (np.narray / torch.tensor): Image array/ tensor which will be sent as an image grid
            make_grid_kargs: Key word arguments for the torchvision make grid method
        """

        caption = ""
        if self.exp_name is not None:
            caption += self.exp_name + "  "
        if name is not None:
            caption += name + "  "

        if isinstance(image_array, np.ndarray):
            image_array = torch.from_numpy(image_array)

        image_array = image_array.cpu()
        grid = torchvision.utils.make_grid(image_array, nrow=nrow, padding=padding, pad_value=pad_value,
                                           normalize=normalize, range=range, scale_each=scale_each)
        ndarr = grid.mul(255).clamp(0, 255).byte().permute(1, 2, 0).numpy()
        buf = get_image_as_buffered_file(ndarr)

        try:
            self.bot.send_photo(chat_id=self.chat_id, photo=buf, caption=caption)
        except Exception as e:
            print("Could not send image_grid to telegram")

    def show_value(self, value, name, counter=None, tag=None, *args, **kwargs):
        """
        Sends a value to a chat using an existing Telegram bot.

        Args:
            value: Value to be plotted sent to the chat.
            name: Name for the plot.
            counter: Optional counter to be sent in conjunction with the value.
            tag: Tag to be used as a label for the plot.
        """

        caption = ""
        if self.exp_name is not None:
            caption += self.exp_name + "  "
        if name is not None:
            caption += name + "  "

        figure = super().show_value(value, name, counter, tag, show=False)
        buf = get_image_as_buffered_file(figure)

        try:
            self.bot.send_photo(chat_id=self.chat_id, photo=buf, caption=caption)
        except Exception as e:
            print("Could not send plot to telegram")

    def show_barplot(self, array, name=None, *args, **kwargs):
        """
        Sends a barplot to a chat using an existing Telegram bot.

        Args:
            array: array of shape NxM where N is the number of rows and M is the number of elements in the row.
            name: The name of the figure

        """

        caption = ""
        if self.exp_name is not None:
            caption += self.exp_name + "  "
        if name is not None:
            caption += name + "  "

        figure = super().show_barplot(array, name, *args, **kwargs)
        buf = get_image_as_buffered_file(figure)

        try:
            self.bot.send_photo(chat_id=self.chat_id, photo=buf, caption=caption)
        except Exception as e:
            print("Could not send plot to telegram")

    def show_lineplot(self, y_vals, x_vals=None, name=None, *args, **kwargs):
        """
        Sends a lineplot to a chat using an existing Telegram bot.

        Args:
            y_vals: Array of shape MxN , where M is the number of points and N is the number of different line
            x_vals: Has to have the same shape as Y: MxN. For each point in Y it gives the corresponding X value (if
            not set the points are assumed to be equally distributed in the interval [0, 1] )
            name: The name of the figure

        """

        caption = ""
        if self.exp_name is not None:
            caption += self.exp_name + "  "
        if name is not None:
            caption += name + "  "

        figure = super().show_lineplot(y_vals, x_vals, name, show=False, *args, **kwargs)
        buf = get_image_as_buffered_file(figure)

        try:
            self.bot.send_photo(chat_id=self.chat_id, photo=buf, caption=caption)
        except Exception as e:
            print("Could not send plot to telegram")

    def show_scatterplot(self, array, name=None, *args, **kwargs):
        """
        Sends a scatterplot to a chat using an existing Telegram bot.

        Args:
            array: A 2d array with size N x dim, where each element i \in N at X[i] results in a a 2d (if dim = 2)/
            3d (if dim = 3) point.
            name: The name of the figure

        """

        caption = ""
        if self.exp_name is not None:
            caption += self.exp_name + "  "
        if name is not None:
            caption += name + "  "

        figure = super().show_scatterplot(array, name, *args, **kwargs)
        buf = get_image_as_buffered_file(figure)

        try:
            self.bot.send_photo(chat_id=self.chat_id, photo=buf, caption=caption)
        except Exception as e:
            print("Could not send plot to telegram")

    def show_piechart(self, array, name=None, *args, **kwargs):
        """
        Sends a piechart to a chat using an existing Telegram bot.

        Args:
            array: Array of positive integers. Each integer will be presented as a part of the pie (with the total
            as the sum of all integers)
            name: The name of the figure

        """

        caption = ""
        if self.exp_name is not None:
            caption += self.exp_name + "  "
        if name is not None:
            caption += name + "  "

        figure = super().show_piechart(array, name,  *args, **kwargs)
        buf = get_image_as_buffered_file(figure)

        try:
            self.bot.send_photo(chat_id=self.chat_id, photo=buf, caption=caption)
        except Exception as e:
            print("Could not send plot to telegram")

    def print(self, text, *args, **kwargs):
        """Just calls show_text()"""
        self.show_text(text, *args, **kwargs)



import os
import tempfile
import unittest
import shutil
import time

import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
from scipy import misc

from trixi.logger.experiment.experimentlogger import ExperimentLogger
from trixi.logger.tensorboard.pytorchtensorboardxlogger import PytorchTensorboardXLogger
from trixi.util.config import Config


class TestPytorchTensorboardXLogger(unittest.TestCase):

    def setUp(self):
        self.test_dir = tempfile.gettempdir()
        self.logger = PytorchTensorboardXLogger(self.test_dir)

    def tearDown(self):
        self.logger.close()
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir, ignore_errors=True)

    def test_show_text(self):
        self.logger.show_text("Test")

    def test_show_image(self):
        image = np.random.random_sample((3, 128, 128))
        self.logger.show_image(image)

    def test_show_images(self):
        images = np.random.random_sample((16, 3, 128, 128))
        self.logger.show_images(images)

    def test_show_image_grid(self):
        images = np.random.random_sample((16, 3, 128, 128))
        self.logger.show_image_grid(images)

    def test_show_barplot(self):
        tensor = np.random.random_sample(5)
        self.logger.show_barplot(tensor)

    def test_show_lineplot(self):
        x = [0, 1, 2, 3, 4, 5]
        y = np.random.random_sample(6)
        self.logger.show_lineplot(y, x, name="lineplot")

    def test_show_piechart(self):
        array = np.random.random_sample(5)
        self.logger.show_piechart(array, name="piechart")

    def test_show_scatterplot(self):
        array = np.random.random_sample((5, 2))
        self.logger.show_scatterplot(array, name="scatterplot")

    def test_show_value(self):
        val = np.random.random_sample(1)
        self.logger.show_value(val, "value")
        val = np.random.random_sample(1)
        self.logger.show_value(val, "value")
        val = np.random.random_sample(1)
        self.logger.show_value(val, "value", counter=4)
        
        val = np.random.random_sample(1)
        self.logger.show_value(val, "value1", tag="xD")
        val = np.random.random_sample(1)
        self.logger.show_value(val, "value2", tag="xD")
        val = np.random.random_sample(1)
        self.logger.show_value(val, "value1", counter=4, tag="xD")

    def test_show_pr_curve(self):
        self.logger.show_pr_curve(np.random.rand(100), np.random.randint(2, size=100))

    def test_show_embedding(self):
        import torch
        label_img = torch.rand(100, 3, 10, 32)
        for i in range(100):
            label_img[i]*=i/100.0
        self.logger.show_embedding(torch.randn(100, 5), label_img=label_img)

    def test_show_histogram(self):
        self.logger.show_histogram(np.random.rand(100))
        


if __name__ == '__main__':
    unittest.main()



from trixi.logger.experiment.experimentlogger import ExperimentLogger
try:
    from trixi.logger.experiment.pytorchexperimentlogger import PytorchExperimentLogger
except ImportError as e:
    import warnings
    warnings.warn(ImportWarning("Could not import Pytorch related modules:\n%s"
        % e.msg))



from collections import defaultdict

from trixi.logger.abstractlogger import AbstractLogger


def create_function(self, sub_methods):
    def surrogate_fn(*args, **kwargs):

        for sub_method in sub_methods:

            try:

                method_cntr = self.log_methods_cntr[sub_method]
                method_freq = self.log_methods_freq[sub_method]

                if method_freq is None or method_freq == 0:
                    continue

                use_name = False
                if "name" in kwargs and not ("ignore_name_in_args" in kwargs and kwargs["ignore_name_in_args"] is True):
                    name_id = kwargs["name"]
                    if "tag" in kwargs:
                        name_id = name_id + kwargs["tag"]
                    method_cntr = self.log_methods_name_cntr[sub_method][name_id]
                    use_name = True

                if "log_all" in kwargs and kwargs["log_all"] is True:
                    sub_method(*args, **kwargs)
                    continue

                if method_cntr % method_freq == 0:
                    sub_method(*args, **kwargs)

                elif "same_as_last" in kwargs and kwargs["same_as_last"] is True:
                    if method_cntr % method_freq == 1:
                        sub_method(*args, **kwargs)
                    kwargs["do_not_increase"] = True

                if use_name:
                    self.log_methods_name_cntr[sub_method][kwargs["name"]] += 1
                elif "do_not_increase" not in kwargs or ("do_not_increase" in kwargs and kwargs["do_not_increase"] is
                                                         False):
                    self.log_methods_cntr[sub_method] += 1

            except Exception as e:
                print("a combi logger method failed: ", str(sub_method))

    return surrogate_fn


class CombinedLogger(object):
    """
    A Logger which can combine all other logger and if called calls all the sub loggers

    """

    def __init__(self, *loggers):
        """
        Initializes a new combined logger with a list of logger and their logging frequencies.

        By default it initalized a counter for each logger method, and on each call it increases the counter (and if the
        counter is a multiple of the frequency it logs to the given logger).

        Furthermore you can use the same_as_last argument in all methods to repeat the frequency_counter from the
        previous call. If you set the do_not_increase attribute to True, it wont increase the counter, it has been
        called 10 times and the frequency is 10, it will log, and also log in the next round, since the counter is
        still 10 and was not increase. Use can use the log_all attribute, it you want to ignore the frequencies and
        simply log to all loggers. By default, if your method has a name attribute, there is a counter for each
        method called with a unique name (so one counter for plot(name="1") and a different counter for plot(name="2")),
        but if you want one counter for the method (despite having a name attribute), you can set the
        ignore_name_in_args to True,

        Args:
            *loggers (list): a list of tuples where each tuple conisist of (logger, frequencies), where logger is
            a given logger and frequencies is the frequency (or rather log each on in frequencies, e.g. if it is 10,
            it logs every ten_th call).
        """

        self.loggers, self.frequencies = zip(*loggers)

        for logger in self.loggers:
            if not isinstance(logger, AbstractLogger):
                raise TypeError("All logger must be subclasses of the abstract visual logger.")
        for freq in self.frequencies:
            if freq is None:
                continue
            elif freq < 0:
                raise ValueError("All frequencies must be positive.")

        self.logger_methods = defaultdict(list)
        self.log_methods_cntr = defaultdict(dict)
        self.log_methods_freq = defaultdict(int)
        self.log_methods_name_cntr = defaultdict(int)

        for logger, freq in zip(self.loggers, self.frequencies):
            if freq is None:
                continue

            logger_vars = [i for i in dir(logger) if not i.startswith("__")]

            for el in logger_vars:
                if hasattr(logger, el) and callable(getattr(logger, el)):
                    self.logger_methods[el].append(getattr(logger, el))
                    self.log_methods_cntr[getattr(logger, el)] = 0
                    self.log_methods_freq[getattr(logger, el)] = freq
                    self.log_methods_name_cntr[getattr(logger, el)] = defaultdict(int)

        for method_name, sub_methods in self.logger_methods.items():
            setattr(self, method_name, create_function(self, sub_methods))



import os
import tempfile
import unittest
import shutil
import time

import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

from trixi.logger.plt.numpyseabornimageplotlogger import NumpySeabornImagePlotLogger
from trixi.logger.visdom import PytorchVisdomLogger
from trixi.logger.visdom.numpyvisdomlogger import start_visdom


class TestNumpySearbonImagePlotLogger(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        super(TestNumpySearbonImagePlotLogger, cls).setUpClass()


    def setUp(self):
        self.logger = NumpySeabornImagePlotLogger()

    def test_show_image(self):
        image = np.random.random_sample((3, 128, 128))
        figure_image = self.logger.show_image(image, "image")
        self.assertTrue(isinstance(figure_image, np.ndarray))

    def test_show_barplot(self):
        tensor = np.random.random_sample(5)
        figure_image = self.logger.show_barplot(tensor, name="barplot")
        self.assertTrue(isinstance(figure_image, np.ndarray))

    def test_show_lineplot(self):
        x = [0, 1, 2, 3, 4, 5]
        y = np.random.random_sample(6)
        figure_image = self.logger.show_lineplot(y, x, name="lineplot1")
        self.assertTrue(isinstance(figure_image, np.ndarray))

    def test_show_piechart(self):
        array = np.random.random_sample(5)
        figure_image = self.logger.show_piechart(array, name="piechart")
        self.assertTrue(isinstance(figure_image, np.ndarray))

    def test_show_scatterplot(self):
        array = np.random.random_sample((5, 2))
        figure_image = self.logger.show_scatterplot(array, name="scatterplot")
        self.assertTrue(isinstance(figure_image, np.ndarray))

    def test_show_value(self):
        val = np.random.random_sample(1)
        figure_image = self.logger.show_value(val, "value")
        self.assertTrue(isinstance(figure_image, np.ndarray))

        val = np.random.random_sample(1)
        figure_image = self.logger.show_value(val, "value")
        self.assertTrue(isinstance(figure_image, np.ndarray))

        val = np.random.random_sample(1)
        figure_image = self.logger.show_value(val, "value", counter=4)
        self.assertTrue(isinstance(figure_image, np.ndarray))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json
from copy import deepcopy

from trixi.util.util import ModuleMultiTypeDecoder, ModuleMultiTypeEncoder


class Config(dict):
    """
    Config is the main object used to store configurations. As a rule of thumb, anything you might
    want to change in your experiment should go into the Config. It's basically a :class:`dict`,
    but vastly more powerful. Key features are

        - Access keys as attributes
            Config["a"]["b"]["c"] is the same as Config.a.b.c.
            Can also be used for setting if the second to last key exists. Only works for keys that
            conform with Python syntax (Config.myattr-1 is not allowed).
        - Advanced de-/serialization
            Using specialized JSON encoders and decoders, almost anything can be serialized and
            deserialized. This includes types, functions (except lambdas) and modules. For example,
            you could have something like::

                c = Config(model=MyModel)
                c.dump("somewhere")

            and end up with a JSON file that looks like this::

                {
                    "model": "__type__(your.model.module.MyModel)"
                }

            and vice versa. We use double underscores and parentheses for serialization,
            so it's probably a good idea to not use this pattern for other stuff!
        - Automatic CLI exposure
            If desired, the Config will create an ArgumentParser that contains all keys in the
            Config as arguments in the form "- - key", so you can run your experiment from the command
            line and manually overwrite certain values. Deeper levels are also accessible via dot
            notation "- - key_with_dict_value.inner_key".
        - Comparison
            Compare any number of Configs and get a new Config containing only the values that
            differ among input Configs.

    Args:
        file_ (str): Load Config from this file.
        config (Config): Update with values from this Config (can be combined with :attr:`file_`).
            Will by default only make shallow copies, see :attr:`deep`.
        update_from_argv (bool): Update values from argv. Will automatically expose keys to the
            CLI as '- - key'.
        deep (bool): Make deep copies if :attr:`config` is given.

    """

    def __init__(self, file_=None, config=None, update_from_argv=False, deep=False, **kwargs):

        super(Config, self).__init__()

        # the following allows us to access keys as attributes if syntax permits
        # config["a"] = 1 -> config.a -> 1
        # config["a-b"] = 2 -> config.a-b (not possible)
        # this is purely for convenience
        self.__dict__ = self

        if file_ is not None:
            self.load(file_)

        if config is not None:
            if deep:
                # convert config to Config (in case it's just a dict)
                # and get deepcopy
                config = Config(config=config, update_from_argv=False, deep=False)
                config = config.deepcopy()
            self.update(config, deep=False)

        if len(kwargs) >= 1:
            if deep:
                kwargs = Config(config=kwargs, update_from_argv=False, deep=False)
                kwargs = kwargs.deepcopy()
            self.update(kwargs, deep=False)

        if update_from_argv:
            update_from_sys_argv(self)

    def update(self, dict_like, deep=False, ignore=None, allow_dict_overwrite=True):
        """Update entries in the Config.

        Args:
            dict_like (dict or derivative thereof): Update source.
            deep (bool): Make deep copies of all references in the source.
            ignore (iterable): Iterable of keys to ignore in update.
            allow_dict_overwrite (bool): Allow overwriting with dict.
                Regular dicts only update on the highest level while we recurse
                and merge Configs. This flag decides whether it is possible to
                overwrite a 'regular' value with a dict/Config at lower levels.
                See examples for an illustration of the difference


        Examples:
            The following illustrates the update behaviour if
            :obj:allow_dict_overwrite is active. If it isn't, an AttributeError
            would be raised, originating from trying to update "string"::

                config1 = Config(config={
                    "lvl0": {
                        "lvl1": "string",
                        "something": "else"
                    }
                })

                config2 = Config(config={
                    "lvl0": {
                        "lvl1": {
                            "lvl2": "string"
                        }
                    }
                })

                config1.update(config2, allow_dict_overwrite=True)

                >>>config1
                {
                    "lvl0": {
                        "lvl1": {
                            "lvl2": "string"
                        },
                        "something": "else"
                    }
                }

        """

        if ignore is None:
            ignore = ()

        if deep:
            update_config = Config(config=dict_like, deep=True)
            self.update(update_config,
                        deep=False,
                        ignore=ignore,
                        allow_dict_overwrite=allow_dict_overwrite)

        else:
            for key, value in dict_like.items():
                if key in ignore:
                    continue
                if key in self and isinstance(value, dict):
                    try:
                        self[key].update(value,
                                         deep=False,
                                         ignore=ignore,
                                         allow_dict_overwrite=allow_dict_overwrite)
                    except AttributeError as ae:
                        if allow_dict_overwrite:
                            self[key] = value
                        else:
                            raise ae
                else:
                    self[key] = value

    def deepupdate(self, dict_like, ignore=None, allow_dict_overwrite=True):
        """Identical to :meth:`update` with `deep=True`.

        Args:
            dict_like (dict or derivative thereof): Update source.
            ignore (iterable): Iterable of keys to ignore in update.
            allow_dict_overwrite (bool): Allow overwriting with dict.
                Regular dicts only update on the highest level while we recurse
                and merge Configs. This flag decides whether it is possible to
                overwrite a 'regular' value with a dict/Config at lower levels.
                See examples for an illustration of the difference

        """
        self.update(dict_like,
                    deep=True,
                    ignore=ignore,
                    allow_dict_overwrite=allow_dict_overwrite)

    def __setattr__(self, key, value):
        """Modified to automatically convert `dict` to Config."""

        if type(value) == dict:
            new_config = Config()
            new_config.update(value, deep=False)
            super(Config, self).__setattr__(key, new_config)
        else:
            super(Config, self).__setattr__(key, value)

    def __getitem__(self, key):
        """Allows convenience access to deeper levels using dots to separate
        levels, for example `config["a.b.c"]`.
        """

        if key == "":
            if len(self.keys()) == 1:
                key = list(self.keys())[0]
            else:
                raise KeyError("Empty string only works for single element Configs.")

        if type(key) == str and "." in key:
            superkey = key.split(".")[0]
            subkeys = ".".join(key.split(".")[1:])
            if type(self[superkey]) in (list, tuple):
                try:
                    subkeys = int(subkeys)
                except ValueError:
                    pass
            return self[superkey][subkeys]
        else:
            return super(Config, self).__getitem__(key)

    def __setitem__(self, key, value):
        """Allows convenience access to deeper levels using dots to separate
        levels, for example `config["a.b.c"]`.
        """

        if key == "":
            if len(self.keys()) == 1:
                key = list(self.keys())[0]
            else:
                raise KeyError("Empty string only works for single element Configs.")

        if type(key) == str and "." in key:
            superkey = key.split(".")[0]
            subkeys = ".".join(key.split(".")[1:])
            if superkey != "" and superkey not in self:
                self[superkey] = Config()
            if type(self[superkey]) == list:
                try:
                    subkeys = int(subkeys)
                except ValueError:
                    pass
            self[superkey][subkeys] = value
        elif type(value) == dict:
            super(Config, self).__setitem__(key, Config(config=value))
        else:
            super(Config, self).__setitem__(key, value)

    def set_with_decode(self, key, value, stringify_value=False):
        """Set single value, using :class:`.ModuleMultiTypeDecoder` to interpret
        key and value strings by creating a temporary JSON string.

        Args:
            key (str): Config key.
            value (str): New value key will map to.
            stringify_value (bool): If `True`, will insert the value into the
                temporary JSON as a real string. See examples!

        Examples:
            Example for when you need to set `stringify_value=True`::

                config.set_with_decode("key", "__type__(trixi.util.config.Config)", stringify_value=True)

            Example for when you need to set `stringify_value=False`::

                config.set_with_decode("key", "[1, 2, 3]")

        """

        if type(key) != str:
            # We could encode the key if it's not a string, but for now raise
            raise TypeError("set_with_decode requires string as key.")
        if type(value) != str:
            raise TypeError("set_with_decode requires string as value.")

        dict_str = ''
        depth = 0
        key_split = key.split(".")

        for k in key_split:
            dict_str += "{"
            dict_str += '"{}":'.format(k)
            depth += 1

        if stringify_value:
            dict_str += '"{}"'.format(value)
        else:
            dict_str += "{}".format(value)

        for _ in range(depth):
            dict_str += "}"

        self.loads(dict_str)

    def set_from_string(self, str_, stringify_value=False):
        """Set a value from a single string, separated with "=".
        Uses :meth:´set_with_decode´.

        Args:
            str_ (str): String that looks like "key=value".

        """

        key, value = str_.split("=")
        self.set_with_decode(key, value, stringify_value)

    def update_missing(self, dict_like, deep=False, ignore=None):
        """Recursively insert values that do not yet exist.

        Args:
            dict_like (dict or derivative thereof): Update source.
            deep (bool): Make deep copies of all references in the source.
            ignore (iterable): Iterable of keys to ignore in update.

        """

        for key, value in dict_like.items():

            if key not in self:
                if type(value) == Config:
                    if deep:
                        self[key] = value.deepcopy()
                    else:
                        self[key] = value
                else:
                    if deep:
                        self[key] = deepcopy(value)
                    else:
                        self[key] = value
            else:
                if isinstance(value, dict) and isinstance(self[key], dict):
                    self[key].update_missing(Config(config=value, deep=deep))

    def dump(self, file_, indent=4, separators=(",", ": "), **kwargs):
        """Write config to file using :meth:`json.dump`.

        Args:
            file_ (str or File): Write to this location.
            indent (int): Formatting option.
            separators (iterable): Formatting option.
            **kwargs: Will be passed to :meth:`json.dump`.

        """

        if hasattr(file_, "write"):
            json.dump(self, file_,
                      cls=ModuleMultiTypeEncoder,
                      indent=indent,
                      separators=separators,
                      **kwargs)
        else:
            with open(file_, "w") as file_object:
                json.dump(self, file_object,
                          cls=ModuleMultiTypeEncoder,
                          indent=indent,
                          separators=separators,
                          **kwargs)

    def dumps(self, indent=4, separators=(",", ": "), **kwargs):
        """Get string representation using :meth:`json.dumps`.

        Args:
            indent (int): Formatting option.
            separators (iterable): Formatting option.
            **kwargs: Will be passed to :meth:`json.dumps`.

        """
        return json.dumps(self,
                          cls=ModuleMultiTypeEncoder,
                          indent=indent,
                          separators=separators,
                          **kwargs)

    def load(self, file_, raise_=True, decoder_cls_=ModuleMultiTypeDecoder, **kwargs):
        """Load config from file using :meth:`json.load`.

        Args:
            file_ (str or File): Read from this location.
            raise (bool): Raise errors.
            decoder_cls_ (type): Class that is used to decode JSON string.
            **kwargs: Will be passed to :meth:`json.load`.

        """

        try:
            if hasattr(file_, "read"):
                new_dict = json.load(file_, cls=decoder_cls_, **kwargs)
            else:
                with open(file_, "r") as file_object:
                    new_dict = json.load(file_object, cls=decoder_cls_, **kwargs)
        except Exception as e:
            if raise_:
                raise e

        self.update(new_dict)

    def loads(self, json_str, decoder_cls_=ModuleMultiTypeDecoder, **kwargs):
        """Load config from JSON string using :meth:`json.loads`.

        Args:
            json_str (str): Interpret this string.
            decoder_cls_ (type): Class that is used to decode JSON string.
            **kwargs: Will be passed to :meth:`json.loads`.

        """

        if not json_str.startswith("{"):
            json_str = "{" + json_str
        if not json_str.endswith("}"):
            json_str = json_str + "}"
        new_dict = json.loads(json_str, cls=decoder_cls_, **kwargs)
        self.update(new_dict)

    def hasattr_not_none(self, key):
        if key in self:
            if self[key] is not None:
                return True
        return False

    def contains(self, dict_like):
        """Check whether all items in a dictionary-like object match the ones in
        this Config.

        Args:
            dict_like (dict or derivative thereof): Returns True if this is
                contained in this Config.

        Returns:
            bool: True if dict_like is contained in self, otherwise False.

        """

        dict_like_config = Config(config=dict_like)

        for key, val in dict_like_config.items():

            if key not in self:
                return False
            else:
                if isinstance(val, dict):
                    if not self[key].contains(val):
                        return False
                else:
                    if not self[key] == val:
                        return False

        return True

    def deepcopy(self):
        """Get a deep copy of this Config.

        Returns:
            Config: A deep copy of self.

        """

        def _deepcopy(source, target):
            for key, val in source.items():
                if not isinstance(val, dict):
                    target[key] = deepcopy(val)
                else:
                    target[key] = Config()
                    _deepcopy(source[key], target[key])

        new_config = Config()
        _deepcopy(self, new_config)

        return new_config

    @staticmethod
    def init_objects(config):
        """Returns a new Config with types converted to instances.

        Any value that is a Config and contains a type key will be converted to
        an instance of that type::

            {
                "stuff": "also_stuff",
                "convert_me": {
                    type: {
                        "param": 1,
                        "other_param": 2
                    },
                    "something_else": "hopefully_useless"
                }
            }

        becomes::

            {
                "stuff": "also_stuff",
                "convert_me": type(param=1, other_param=2)
            }

        Note that additional entries can be lost as shown above.

        Args:
            config (Config): New Config will be built from this one

        Returns:
            Config: A new config with instances made from type entries.
        """

        def init_sub_objects(objs):
            if isinstance(objs, dict):
                ret_dict = Config()
                for key, val in objs.items():
                    if isinstance(key, type):
                        init_param = init_sub_objects(val)
                        if isinstance(init_param, dict):
                            init_obj = key(**init_param)
                        elif isinstance(init_param, (list, tuple, set)):
                            init_obj = key(*init_param)
                        else:
                            init_obj = key()
                        return init_obj
                    elif isinstance(val, (dict, list, tuple, set)):
                        ret_dict[key] = init_sub_objects(val)
                    else:
                        ret_dict[key] = val
                return ret_dict
            elif isinstance(objs, (list, tuple, set)):
                orig_type = type(objs)
                ret_list = []
                for el in objs:
                    ret_list.append(init_sub_objects(el))
                return orig_type(ret_list)
            else:
                return objs

        return init_sub_objects(config)

    def __str__(self):
        return self.dumps(sort_keys=True)

    def difference_config(self, *other_configs):
        """Get the difference of this and any number of other configs.
        See :meth:`difference_config_static` for more information.

        Args:
            *other_configs (Config): Compare these configs and self.

        Returns:
            Config: Difference of self and the other configs.

        """
        return self.difference_config_static(self, *other_configs)

    @staticmethod
    def difference_config_static(*configs, only_set=False):
        """Make a Config of all elements that differ between N configs.

        The resulting Config looks like this::

            {
                key: (config1[key], config2[key], ...)
            }

        If the key is missing, None will be inserted. The inputs will not be
        modified.

        Args:
            configs (Config): Any number of Configs
            only_set (bool): If only the set of different values hould be returned or for each config the
            corresponding one

        Returns:
            Config: Possibly empty Config
        """

        difference = dict()
        mmte = ModuleMultiTypeEncoder()

        all_keys = set()
        for config in configs:
            all_keys.update(set(config.keys()))

        for key in all_keys:

            current_values = []
            all_equal = True
            all_configs = True

            for config in configs:

                if key not in config:
                    all_equal = False
                    all_configs = False
                    current_values.append(None)
                else:
                    current_values.append(mmte._encode(config[key]))

                if len(current_values) >= 2:
                    if current_values[-1] != current_values[-2]:
                        all_equal = False

                if type(current_values[-1]) != Config:
                    all_configs = False

            if not all_equal:

                if not all_configs:
                    if not only_set:
                        difference[key] = tuple(current_values)
                    else:
                        difference[key] = tuple(set(current_values))
                else:
                    difference[key] = Config.difference_config_static(*current_values, only_set=only_set)

        return difference

    def flat(self, keep_lists=True, max_split_size=10):
        """Returns a flattened version of the Config as dict.

        Nested Configs and lists will be replaced by concatenated keys like so::

            {
                "a": 1,
                "b": [2, 3],
                "c": {
                    "x": 4,
                    "y": {
                        "z": 5
                    }
                },
                "d": (6, 7)
            }

        Becomes::

            {
                "a": 1,
                "b": [2, 3], # if keep_lists is True
                "b.0": 2,
                "b.1": 3,
                "c.x": 4,
                "c.y.z": 5,
                "d": (6, 7)
            }

        We return a dict because dots are disallowed within Config keys.

        Args:
            keep_lists: Keeps list along with unpacked values
            max_split_size: List longer than this will not be unpacked

        Returns:
            dict: A flattened version of self
        """

        def flat_(obj):
            def items():
                for key, val in obj.items():
                    if isinstance(val, dict) and type(key) == str:
                        intermediate_dict = {}
                        for subkey, subval in flat_(val).items():
                            if type(subkey) == str:
                                yield key + "." + subkey, subval
                            else:
                                intermediate_dict[subkey] = subval
                        if len(intermediate_dict) > 0:
                            yield key, intermediate_dict
                    elif isinstance(val, (list, tuple)):
                        keep_this = keep_lists or type(key) != str
                        if max_split_size not in (None, False) and len(val) > max_split_size:
                            keep_this = True
                        if keep_this:
                            yield key, val
                        if max_split_size in (None, False) or len(val) <= max_split_size:
                            for i, subval in enumerate(val):
                                yield key + "." + str(i), subval
                    else:
                        yield key, val

            return dict(items())

        return flat_(self)

    def to_cmd_args_str(self):
        """Create a string representing what one would need to pass to the
        command line. Does not yet use JSON encoding!

        Returns:
            str: Command line string

        """

        c_flat = self.flat()

        str_list = []
        for key, val in c_flat.items():

            if isinstance(val, (list, tuple)):
                vals = [str(v) for v in val]
                val_str = " ".join(vals)
            else:
                val_str = str(val)
            str_list.append("--{} {}".format(key, val_str))

        return "  ".join(str_list)


def update_from_sys_argv(config, warn=False):
    """Updates Config with the arguments passed as args when running the
    program. Keys will be converted to command line options, then matching
    options in `sys.argv` will be used to update the Config.

    Args:
        config (Config): Update this Config.
        warn (bool): Raise warnings if there are unknown options. Turn this on
            if you don't use any :class:`argparse.ArgumentParser` after to
            check for possible errors.

    """

    import sys
    import argparse
    import warnings

    def str2bool(v):
        if v.lower() in ('yes', 'true', 't', 'y', '1'):
            return True
        elif v.lower() in ('no', 'false', 'f', 'n', '0'):
            return False
        else:
            raise argparse.ArgumentTypeError('Boolean value expected.')

    if len(sys.argv) > 1:

        parser = argparse.ArgumentParser(allow_abbrev=False)
        encoder = ModuleMultiTypeEncoder()
        decoder = ModuleMultiTypeDecoder()

        config_flat = config.flat()
        for key, val in config_flat.items():
            name = "--{}".format(key)
            if val is None:
                parser.add_argument(name)
            else:
                if type(val) == bool:
                    parser.add_argument(name, type=str2bool, default=val)
                elif isinstance(val, (list, tuple)):
                    if len(val) > 0 and type(val[0]) != type:
                        parser.add_argument(name, nargs='+', type=type(val[0]), default=val)
                    else:
                        parser.add_argument(name, nargs='+', default=val)
                else:
                    if type(val) == type:
                        val = encoder._encode(val)
                    parser.add_argument(name, type=type(val), default=val)

        # parse args
        param, unknown = parser.parse_known_args()
        param = vars(param)

        if len(unknown) > 0 and warn:
            warnings.warn("Called with unknown arguments: {}".format(unknown), RuntimeWarning)

        # calc diff between configs
        diff_keys = list(Config.difference_config_static(param, config_flat).keys())

        # convert type args
        ignore_ = []
        for key, val in param.items():
            if val in ("none", "None"):
                param[key] = None
            if type(config_flat[key]) == type:
                if isinstance(val, str):
                    val = val.replace("\'", "")
                    val = val.replace("\"", "")
                param[key] = decoder._decode(val)
            try:
                key_split = key.split(".")
                list_object, _ = ".".join(key_split[:-1]), int(key_split[-1])
                if "--" + list_object in sys.argv:
                    ignore_.append(key)
            except ValueError as ve:
                pass
        for i in ignore_:
            del param[i]

        ### Delete not changed entries
        param_keys = list(param.keys())
        for i in param_keys:
            if i not in diff_keys and i in param:
                del param[i]

        # update dict
        config.update(param)



import os
import tempfile
import unittest
import shutil
import time

import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
from trixi.logger.visdom import PytorchVisdomLogger
from trixi.logger.visdom.numpyvisdomlogger import start_visdom


class TestPytorchVisdomLogger(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        super(TestPytorchVisdomLogger, cls).setUpClass()
        try:
            start_visdom()
        except:
            print("Could not start visdom, it might be already running.")

    def setUp(self):
        self.visdomLogger = PytorchVisdomLogger()

    def test_show_image(self):
        image = np.random.random_sample((3, 128, 128))
        tensor = torch.from_numpy(image)
        self.visdomLogger._NumpyVisdomLogger__show_image(tensor.numpy(), "image")

    def test_show_images(self):
        images = np.random.random_sample((4, 3, 128, 128))
        tensors = torch.from_numpy(images)
        self.visdomLogger._NumpyVisdomLogger__show_images(tensors.numpy(), "image")

    def test_show_image_grid(self):
        images = np.random.random_sample((4, 3, 128, 128))
        tensor = torch.from_numpy(images)
        self.visdomLogger._PytorchVisdomLogger__show_image_grid(tensor, "image_grid")

    def test_show_image_grid_heatmap(self):
        images = np.random.random_sample((4, 3, 128, 128))
        self.visdomLogger._PytorchVisdomLogger__show_image_grid_heatmap(images, name="image_grid_heatmap")

    def test_show_barplot(self):
        tensor = torch.from_numpy(np.random.random_sample(5))
        self.visdomLogger.show_barplot(tensor, name="barplot")
        self.visdomLogger._NumpyVisdomLogger__show_barplot(tensor.numpy(), name="barplot")

    def test_show_lineplot(self):
        x = [0, 1, 2, 3, 4, 5]
        y = np.random.random_sample(6)
        self.visdomLogger.show_lineplot(y, x, name="lineplot1")
        self.visdomLogger._NumpyVisdomLogger__show_lineplot(y, x, name="lineplot1")

    def test_show_piechart(self):
        array = torch.from_numpy(np.random.random_sample(5))
        self.visdomLogger.show_piechart(array, name="piechart")
        self.visdomLogger._NumpyVisdomLogger__show_piechart(array, name="piechart")

    def test_show_scatterplot(self):
        array = torch.from_numpy(np.random.random_sample((5, 2)))
        self.visdomLogger.show_scatterplot(array, name="scatterplot")
        self.visdomLogger._NumpyVisdomLogger__show_scatterplot(array.numpy(), name="scatterplot")

    def test_show_value(self):
        val = torch.from_numpy(np.random.random_sample(1))
        self.visdomLogger.show_value(val, "value")
        self.visdomLogger._NumpyVisdomLogger__show_value(val.numpy(), "value")

        val = torch.from_numpy(np.random.random_sample(1))
        self.visdomLogger.show_value(val, "value")

        val = torch.from_numpy(np.random.random_sample(1))
        self.visdomLogger.show_value(val, "value", counter=4)

    def test_show_text(self):
        text = "\nTest 4 fun: zD ;-D 0o"
        self.visdomLogger.show_text(text)
        self.visdomLogger._NumpyVisdomLogger__show_text(text)

    def test_get_roc_curve(self):
        array = np.random.random_sample(100)
        labels = np.random.choice((0, 1), 100)

        self.visdomLogger.show_roc_curve(array, labels, name="roc")

    def test_get_pr_curve(self):
        array = np.random.random_sample(100)
        labels = np.random.choice((0, 1), 100)

        self.visdomLogger.show_roc_curve(array, labels, name="pr")

    def test_get_classification_metric(self):
        array = np.random.random_sample(100)
        labels = np.random.choice((0, 1), 100)

        self.visdomLogger.show_classification_metrics(array, labels, metric=("roc-auc", "pr-score"),
                                                      name="classification-metrics")

    def test_show_image_gradient(self):
        net = Net()
        random_input = torch.from_numpy(np.random.randn(28 * 28).reshape((1, 1, 28, 28))).float()
        fake_labels = torch.from_numpy(np.array([2])).long()
        criterion = torch.nn.CrossEntropyLoss()

        err_fn = lambda x: criterion(x, fake_labels)

        self.visdomLogger.show_image_gradient(name="grads-vanilla", model=net, inpt=random_input, err_fn=err_fn,
                                              grad_type="vanilla")
        time.sleep(1)

        self.visdomLogger.show_image_gradient(name="grads-svanilla", model=net, inpt=random_input, err_fn=err_fn,
                                              grad_type="smooth-vanilla")
        time.sleep(1)

        self.visdomLogger.show_image_gradient(name="grads-guided", model=net, inpt=random_input, err_fn=err_fn,
                                              grad_type="guided")
        time.sleep(1)

        self.visdomLogger.show_image_gradient(name="grads-sguided", model=net, inpt=random_input, err_fn=err_fn,
                                              grad_type="smooth-guided")
        time.sleep(1)

    def test_plot_model_structure(self):
        net = Net()
        self.visdomLogger.plot_model_structure(net, [(1, 1, 28, 28)])


    def test_plot_model_statistics(self):
        net = Net()
        self.visdomLogger.plot_model_statistics(net, plot_grad=False)
        self.visdomLogger.plot_model_statistics(net, plot_grad=True)

    def test_show_embedding(self):
        array = torch.from_numpy(np.random.random_sample((100, 100)))
        self.visdomLogger.show_embedding(array, method="tsne")
        self.visdomLogger.show_embedding(array, method="umap")


class Net(nn.Module):
    """
    Small network to test save/load functionality
    """

    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)


if __name__ == '__main__':
    unittest.main()



import time


class Experiment(object):
    """
    An abstract Experiment which can be run for a number of epochs.

    The basic life cycle of an experiment is::

        setup()
        prepare()

        for epoch in n_epochs:
            train()
            validate()

        end()

    The reason there is both :meth:`.setup` and :meth:`.prepare` is that internally there is also
    a :meth:`._setup_internal` method for hidden magic in classes that inherit from this. For
    example, the :class:`trixi.experiment.pytorchexperiment.PytorchExperiment` uses this to restore checkpoints. Think
    if :meth:`.setup` as an :meth:`.__init__` that is only called when the Experiment is actually
    asked to do anything. Then use :meth:`.prepare` to modify the fully instantiated Experiment if
    you need to.

    To write a new Experiment simply inherit the Experiment class and overwrite the methods.
    You can then start your Experiment calling :meth:`.run`

    In Addition the Experiment also has a test function. If you call the :meth:`.run_test` method it
    will call the :meth:`.test` and :meth:`.end_test` method internally (and if you give the
    parameter setup = True in run_test is will again call :meth:`.setup` and :meth:`.prepare` ).

    Each Experiment also has its current state in  :attr:`_exp_state`, its start time in
    :attr:`_time_start`, its end time in :attr:`_time_end` and the current epoch index in
    :attr:`_epoch_idx`

    Args:
        n_epochs (int): The number of epochs in the Experiment (how often the train and validate
            method will be called)

    """

    def __init__(self, n_epochs=0):

        self.n_epochs = n_epochs
        self._exp_state = "Preparing"
        self._time_start = ""
        self._time_end = ""
        self._epoch_idx = 0

    def run(self, setup=True):
        """
        This method runs the Experiment. It runs through the basic lifecycle of an Experiment::

            setup()
            prepare()

            for epoch in n_epochs:
                train()
                validate()

            end()

        """

        try:
            self._time_start = time.strftime("%y-%m-%d_%H:%M:%S", time.localtime(time.time()))
            self._time_end = ""

            if setup:
                self.setup()
                self._setup_internal()
            self.prepare()

            self._exp_state = "Started"
            self._start_internal()
            print("Experiment started.")

            for epoch in range(self._epoch_idx, self.n_epochs):
                self.train(epoch=epoch)
                self.validate(epoch=epoch)
                self._end_epoch_internal(epoch=epoch)
                self._epoch_idx += 1

            self._exp_state = "Trained"
            print("Training complete.")

            self.end()
            self._end_internal()
            self._exp_state = "Ended"
            print("Experiment ended.")

            self._time_end = time.strftime("%y-%m-%d_%H:%M:%S", time.localtime(time.time()))

        except Exception as e:

            # run_error = e
            # run_error_traceback = traceback.format_tb(e.__traceback__)
            self._exp_state = "Error"
            self.process_err(e)
            self._time_end = time.strftime("%y-%m-%d_%H:%M:%S", time.localtime(time.time()))

            raise e

    def run_test(self, setup=True):
        """
        This method runs the Experiment.

        The test consist of an optional setup and then calls the :meth:`.test` and :meth:`.end_test`.

        Args:
            setup: If True it will execute the :meth:`.setup` and :meth:`.prepare` function similar to the run method
                before calling :meth:`.test`.

        """

        try:

            if setup:
                self.setup()
                self._setup_internal()
                self.prepare()

            self._exp_state = "Testing"
            print("Start test.")

            self.test()
            self.end_test()
            self._exp_state = "Tested"

            self._end_test_internal()

            print("Testing complete.")

        except Exception as e:

            # run_error = e
            # run_error_traceback = traceback.format_tb(e.__traceback__)
            self._exp_state = "Error"
            self.process_err(e)

            raise e

    def setup(self):
        """Is called at the beginning of each Experiment run to setup the basic components needed for a run"""
        pass

    def train(self, epoch):
        """
        The training part of the Experiment, it is called once for each epoch

        Args:
            epoch (int): The current epoch the train method is called in

        """
        pass

    def validate(self, epoch):
        """
        The evaluation/validation part of the Experiment, it is called once for each epoch (after the training
        part)

        Args:
            epoch (int):  The current epoch the validate method is called in

        """
        pass

    def test(self):
        """The testing part of the Experiment"""
        pass

    def process_err(self, e):
        """
        This method is called if an error occurs during the execution of an experiment

        Args:
            e (Exception): The exception which was raised during the experiment life cycle

        """
        pass

    def _setup_internal(self):
        pass

    def _end_epoch_internal(self, epoch):
        pass

    def end(self):
        """Is called at the end of each experiment"""
        pass

    def end_test(self):
        """Is called at the end of each experiment test"""
        pass

    def prepare(self):
        """This method is called directly before the experiment training starts"""
        pass

    def _start_internal(self):
        pass

    def _end_internal(self):
        pass

    def _end_test_internal(self):
        pass



import os
import re

from setuptools import setup, find_packages


def resolve_requirements(file):
    requirements = []
    with open(file) as f:
        req = f.read().splitlines()
        for r in req:
            if r.startswith("-r"):
                requirements += resolve_requirements(os.path.join(os.path.dirname(file), r.split(" ")[1]))
            else:
                requirements.append(r)
    return requirements


def read_file(file):
    with open(file) as f:
        content = f.read()
    return content


def find_version(file):
    content = read_file(file)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", content, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


required = resolve_requirements(os.path.join(os.path.dirname(__file__), 'requirements_full.txt'))
readme = read_file(os.path.join(os.path.dirname(__file__), "Readme.md"))
license = read_file(os.path.join(os.path.dirname(__file__), "LICENSE"))
version = find_version(os.path.join(os.path.dirname(__file__), "trixi", "__init__.py"))


setup(name='trixi',
      version=version,
      description='Manage your machine learning experiments with trixi - modular, reproducible, high fashion',
      long_description=readme,
      long_description_content_type="text/markdown",
      url='https://github.com/MIC-DKFZ/trixi',
      author='Medical Image Computing Group, DKFZ',
      author_email='mic@dkfz-heidelberg.de',
      license=license,
      packages=find_packages(),
      install_requires=required,
      zip_safe=True,
      entry_points={
          'console_scripts': ['trixi-browser=trixi.experiment_browser.browser:start_browser'],
      },
      include_package_data=True
      )



import numpy as np
import torch
import torchvision
from slackclient import SlackClient

from trixi.util.util import get_image_as_buffered_file

from trixi.logger.plt.numpyseabornimageplotlogger import NumpySeabornImagePlotLogger


class SlackMessageLogger(NumpySeabornImagePlotLogger):
    """
    Slack logger, inherits the NumpySeabornImagePlotLogger and sends plots/logs to a chat via a Slack bot.
    """

    @staticmethod
    def find_uid_for_email(slack_client, email):
        """
        Returns the slack user id for a given email

        Args:
            slack_client: Slack client (already authorized)
            email: Workspace email address to get the user id for

        Returns:
            Slack workspace user id

        """

        user_list = slack_client.api_call("users.list").get('members', [])
        user_id = ""

        for user in user_list:
            if user.get('profile', {}).get('email', '') == email:
                user_id = user['id']
                break

        return user_id

    @staticmethod
    def find_cid_for_user(slack_client, uid):
        """
        Returns the chat/channel id for a direct message of the bot with the given user

        Args:
            slack_client: Slack client (already authorized)
            uid: User id of the user

        Returns:
            chat/channel id for a direct message
        """

        direct_conv = slack_client.api_call(
            "conversations.open",
            users=[uid]
        )

        c_id = direct_conv.get('channel', {}).get('id', None)

        return c_id

    def __init__(self, token, user_email, exp_name=None, *args, **kwargs):
        """
        Creates a new NumpySeabornImagePlotLogger object.

        Args:
            token (str): The Bot User OAuth Access Token of the slack bot used.
            user_email (str): The user email in the workspace for the chat between the user and the slack bot.
            exp_name: name of the experiment, always used as message prefix
        """
        super(SlackMessageLogger, self).__init__(**kwargs)

        self.token = token
        self.user_email = user_email
        self.slack_client = SlackClient(token)
        self.uid = self.find_uid_for_email(self.slack_client, self.user_email)
        self.cid = self.find_cid_for_user(self.slack_client, self.uid)
        self.exp_name = exp_name

        self.ts_dict = {}

    def process_params(self, f, *args, **kwargs):
        """
        Inherited "decorator": convert PyTorch variables and Tensors to numpy arrays
        """

        # convert args
        args = (a.detach().cpu().numpy() if torch.is_tensor(a) else a for a in args)

        # convert kwargs
        for key, data in kwargs.items():
            if torch.is_tensor(data):
                kwargs[key] = data.detach().cpu().numpy()

        return f(self, *args, **kwargs)

    def send_message(self, message="", file=None):
        """
        Sends a message and a file if one is given

        Args:
            message: Message to be sent
            file:  File to be sent

        Returns:
            The timestamp (ts) of the message

        """
        if file is None:
            ret_val = self.slack_client.api_call(
                "chat.postMessage",
                channel=self.cid,
                text=message
            )
            return ret_val.get('ts', '')
        else:
            ret_val = self.slack_client.api_call(
                "files.upload",
                channels=self.cid,
                file=file,
                title=message
            )
            try:
                ts = list(ret_val['file']['shares']['private'].values())[0][0]['ts']
            except Exception as e:
                ts = ""
            return ts

    def delete_message(self, ts):
        """
        Deletes a direct message from the bot with the given timestamp (ts)

        Args:
            ts: Time stamp the message was sent

        """
        self.slack_client.api_call(
          "chat.delete",
          channel=self.cid,
          ts=ts
        )

    def show_text(self, text, *args, **kwargs):
        """
        Sends a text to a chat using an existing slack bot.

        Args:
            text (str): Text message to be sent to the bot.
        """
        if self.exp_name is not None:
            text = self.exp_name + ":\n" + text
        try:
            self.send_message(message=text)
        except Exception as e:
            print("Could not send text to slack")

    def show_image(self, image, *args,  **kwargs):
        """
        Sends an image file to a chat using an existing slack bot.

        Args:
            image (str or np array): Path to the image file to be sent to the chat.
        """
        try:
            if isinstance(image, str):
                with open(image, 'rb') as img_file:
                    self.send_message(message=self.exp_name, file=img_file)
            elif isinstance(image, np.ndarray):
                buf = get_image_as_buffered_file(image)
                self.send_message(message=self.exp_name, file=buf)

        except Exception as e:
            print("Could not send image to slack")

    def show_image_grid(self, image_array, name=None, nrow=8, padding=2,
                        normalize=False, range=None, scale_each=False, pad_value=0, delete_last=True,
                        *args, **kwargs):
        """
        Sends an array of images to a chat using  an existing Slack bot. (Requires torch and torchvision)


        Args:
            image_array (np.narray / torch.tensor): Image array/tensor which will be sent as an image grid
            make_grid_kargs: Key word arguments for the torchvision make grid method
            delete_last: If a message with the same name was sent, delete it beforehand
        """

        caption = ""
        if self.exp_name is not None:
            caption += self.exp_name + "  "
        if name is not None:
            caption += name + "  "

        if delete_last and name in self.ts_dict:
            self.delete_message(self.ts_dict[name])

        if isinstance(image_array, np.ndarray):
            image_array = torch.from_numpy(image_array)

        image_array = image_array.cpu()
        grid = torchvision.utils.make_grid(image_array, nrow=nrow, padding=padding, pad_value=pad_value,
                                           normalize=normalize, range=range, scale_each=scale_each)
        ndarr = grid.mul(255).clamp(0, 255).byte().numpy()
        buf = get_image_as_buffered_file(ndarr)
        try:
            ts = self.send_message(message=caption, file=buf)
            self.ts_dict[name] = ts
        except Exception as e:
            print("Could not send image_grid to slack")

    def show_value(self, value, name, counter=None, tag=None, delete_last=True, *args, **kwargs):
        """
        Sends a value to a chat using an existing slack bot.

        Args:
            value: Value to be plotted sent to the chat.
            name: Name for the plot.
            counter: Optional counter to be sent in conjunction with the value.
            tag: Tag to be used as a label for the plot.
            delete_last: If a message with the same name was sent, delete it beforehand
        """

        caption = ""
        if self.exp_name is not None:
            caption += self.exp_name + "  "
        if name is not None:
            caption += name + "  "

        if delete_last and name in self.ts_dict:
            self.delete_message(self.ts_dict[name])

        figure = super().show_value(value=value, name=name, counter=counter, tag=tag)
        buf = get_image_as_buffered_file(figure)
        try:
            ts = self.send_message(message=caption, file=buf)
            self.ts_dict[name] = ts
        except Exception as e:
            print("Could not send plot to slack")

    def show_barplot(self, array, name="barplot", delete_last=True, *args, **kwargs):
        """
        Sends a barplot to a chat using an existing slack bot.

        Args:
            array: array of shape NxM where N is the number of rows and M is the number of elements in the row.
            name: The name of the figure
            delete_last: If a message with the same name was sent, delete it beforehand

        """

        caption = ""
        if self.exp_name is not None:
            caption += self.exp_name + "  "
        if name is not None:
            caption += name + "  "

        if delete_last and name in self.ts_dict:
            self.delete_message(self.ts_dict[name])

        figure = super().show_barplot(array, name, *args, **kwargs)
        buf = get_image_as_buffered_file(figure)
        try:
            ts = self.send_message(message=caption, file=buf)
            self.ts_dict[name] = ts
        except Exception as e:
            print("Could not send plot to slack")

    def show_lineplot(self, y_vals, x_vals=None, name="lineplot", delete_last=True, *args, **kwargs):
        """
        Sends a lineplot to a chat using an existing slack bot.

        Args:
            y_vals: Array of shape MxN , where M is the number of points and N is the number of different line
            x_vals: Has to have the same shape as Y: MxN. For each point in Y it gives the corresponding X value (if
                not set the points are assumed to be equally distributed in the interval [0, 1])
            name: The name of the figure
            delete_last: If a message with the same name was sent, delete it beforehand

        """

        caption = ""
        if self.exp_name is not None:
            caption += self.exp_name + "  "
        if name is not None:
            caption += name + "  "

        if delete_last and name in self.ts_dict:
            self.delete_message(self.ts_dict[name])

        figure = super().show_lineplot(y_vals, x_vals, name, *args, **kwargs)
        buf = get_image_as_buffered_file(figure)
        try:
            ts = self.send_message(message=caption, file=buf)
            self.ts_dict[name] = ts
        except Exception as e:
            print("Could not send plot to slack")

    def show_scatterplot(self, array, name="scatterplot", delete_last=True, *args, **kwargs):
        """
        Sends a scatterplot to a chat using an existing slack bot.

        Args:
            array: An array with size N x dim, where each element i \in N at X[i] results in a 2D
                (if dim = 2) or 3D (if dim = 3) point.
            name: The name of the figure
            delete_last: If a message with the same name was sent, delete it beforehand

        """

        caption = ""
        if self.exp_name is not None:
            caption += self.exp_name + "  "
        if name is not None:
            caption += name + "  "

        if delete_last and name in self.ts_dict:
            self.delete_message(self.ts_dict[name])

        figure = super().show_scatterplot(array, name, *args, **kwargs)
        buf = get_image_as_buffered_file(figure)

        try:
            ts = self.send_message(message=caption, file=buf)
            self.ts_dict[name] = ts
        except Exception as e:
            print("Could not send plot to slack")

    def show_piechart(self, array, name="piechart", delete_last=True, *args, **kwargs):
        """
        Sends a piechart to a chat using an existing slack bot.

        Args:
            array: Array of positive integers. Each integer will be presented as a part of the pie (with the total
                as the sum of all integers)
            name: The name of the figure
            delete_last: If a message with the same name was sent, delete it beforehand

        """

        caption = ""
        if self.exp_name is not None:
            caption += self.exp_name + "  "
        if name is not None:
            caption += name + "  "

        if delete_last and name in self.ts_dict:
            self.delete_message(self.ts_dict[name])

        figure = super().show_piechart(array, name, *args, **kwargs)
        buf = get_image_as_buffered_file(figure)

        try:
            ts = self.send_message(message=caption, file=buf)
            self.ts_dict[name] = ts
        except Exception as e:
            print("Could not send plot to slack")

    def print(self, text, *args, **kwargs):
        """Just calls :meth:`.show_text`"""
        self.show_text(text, *args, **kwargs)



import os

from trixi.logger.abstractlogger import convert_params
from trixi.logger.plt.numpyseabornplotlogger import NumpySeabornPlotLogger
from trixi.util import savefig_and_close


# this is just to turn threaded into non-threaded
def threaded(func):
    return func


class NumpyPlotFileLogger(NumpySeabornPlotLogger):
    """
    NumpyPlotFileLogger is a logger, which can plot/ interpret numpy array as different types (images, lineplots, ...)
    into an image and plot directory. For the plotting it builds up on the NumpySeabornPlotLogger.

    """

    def __init__(self, img_dir, plot_dir, **kwargs):
        """
        Initializes a numpy plot file logger to plot images and plots into an image and plot directory

        Args:
            img_dir: The directory to store images in
            plot_dir: The directory to store plots in
        """
        super(NumpyPlotFileLogger, self).__init__(**kwargs)
        self.img_dir = img_dir
        self.plot_dir = plot_dir

    @convert_params
    def show_image(self, image, name, file_format=".png", *args, **kwargs):
        """
        Method which stores an image as a image file

        Args:
            image: Numpy array-image
            name: file-name
            file_format: output-image file format

        """
        figure = NumpySeabornPlotLogger.show_image(self, image, name, show=False)
        outname = os.path.join(self.img_dir, name) + file_format
        os.makedirs(os.path.dirname(outname), exist_ok=True)
        threaded(savefig_and_close)(figure, outname)

    @convert_params
    def show_value(self, value, name, counter=None, tag=None, file_format=".png", *args, **kwargs):
        """
        Method which logs a value as a line plot

        Args:
            value: Value (y-axis value) you want to display/ plot/ store
            name: Name of the value (will also be the filename if no tag is given)
            counter: counter, which tells the number of the sample (with the same name --> filename) (x-axis value)
            tag: Tag, grouping similar values. Values with the same tag will be plotted in the same plot
            file_format: output-image file format

        Returns:

        """
        figure = NumpySeabornPlotLogger.show_value(self, value, name, counter, tag, show=False)
        if tag is None:
            outname = os.path.join(self.plot_dir, name) + file_format
        else:
            outname = os.path.join(self.plot_dir, tag) + file_format
        os.makedirs(os.path.dirname(outname), exist_ok=True)
        threaded(savefig_and_close)(figure, outname)

    @convert_params
    def show_barplot(self, array, name, file_format=".png", *args, **kwargs):
        """
        Method which creates and stores a barplot

        Args:
            array: Array of values you want to plot
            name: file-name
            file_format: output-image (plot) file format
        """
        figure = NumpySeabornPlotLogger.show_barplot(self, array, name, show=False)
        outname = os.path.join(self.plot_dir, name) + file_format
        os.makedirs(os.path.dirname(outname), exist_ok=True)
        threaded(savefig_and_close)(figure, outname)

    @convert_params
    def show_boxplot(self, array, name, file_format=".png", *args, **kwargs):
        """
        Method which creates and stores a boxplot

        Args:
            array: Array of values you want to plot
            name: file-name
            file_format: output-image (plot) file format
        """
        figure = NumpySeabornPlotLogger.show_boxplot(self, array, name, show=False, **kwargs)
        outname = os.path.join(self.plot_dir, name) + file_format
        os.makedirs(os.path.dirname(outname), exist_ok=True)
        threaded(savefig_and_close)(figure, outname)

    @convert_params
    def show_lineplot(self, y_vals, x_vals, name, file_format=".png", *args, **kwargs):
        """
        Method which creates and stores a lineplot

        Args:
            y_vals: Array of y values
            x_vals: Array of corresponding x-values
            name: file-name
            file_format: output-image (plot) file format
        """
        figure = NumpySeabornPlotLogger.show_lineplot(self, y_vals, x_vals, name, show=False)
        outname = os.path.join(self.plot_dir, name) + file_format
        os.makedirs(os.path.dirname(outname), exist_ok=True)
        threaded(savefig_and_close)(figure, outname)

    @convert_params
    def show_scatterplot(self, array, name, file_format=".png", *args, **kwargs):
        """
        Method which creates and stores a scatter

        Args:
            array: Array of values you want to plot
            name: file-name
            file_format: output-image (plot) file format
        """
        figure = NumpySeabornPlotLogger.show_scatterplot(self, array, name, show=False)
        outname = os.path.join(self.plot_dir, name) + file_format
        os.makedirs(os.path.dirname(outname), exist_ok=True)
        threaded(savefig_and_close)(figure, outname)

    @convert_params
    def show_piechart(self, array, name, file_format=".png", *args, **kwargs):
        """
        Method which creates and stores a piechart

        Args:
            array: Array of values you want to plot
            name: file-name
            file_format: output-image (plot) file format
        """
        figure = NumpySeabornPlotLogger.show_piechart(self, array, name, show=False)
        outname = os.path.join(self.plot_dir, name) + file_format
        os.makedirs(os.path.dirname(outname), exist_ok=True)
        threaded(savefig_and_close)(figure, outname)
